# distutils: sources = yt/utilities/lib/tsearch.c
# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
"""
Oct container




"""


cimport cython
cimport numpy as np

import numpy as np

from libc.math cimport floor

from .oct_visitors cimport (
    NeighbourCellIndexVisitor,
    NeighbourCellVisitor,
    OctPadded,
    StoreIndex,
)
from .selection_routines cimport AlwaysSelector, SelectorObject

ORDER_MAX = 20
_ORDER_MAX = ORDER_MAX

cdef extern from "stdlib.h":
    # NOTE that size_t might not be int
    void *alloca(int)

# Here is the strategy for RAMSES containers:
#   * Read each domain individually, creating *all* octs found in that domain
#     file, even if they reside on other CPUs.
#   * Only allocate octs that reside on >= domain
#   * For all octs, insert into tree, which may require traversing existing
#     octs
#   * Note that this does not allow one component of an ObjectPool (an
#     AllocationContainer) to exactly be a chunk, but it is close.  For IO
#     chunking, we can theoretically examine those octs that live inside a
#     given allocator.

cdef class OctreeContainer:

    def __init__(self, oct_domain_dimensions, domain_left_edge,
                 domain_right_edge, partial_coverage = 0,
                 num_zones = 2):
        # This will just initialize the root mesh octs
        self.nz = num_zones
        self.partial_coverage = partial_coverage
        cdef int i
        for i in range(3):
            self.nn[i] = oct_domain_dimensions[i]
        self.num_domains = 0
        self.level_offset = 0
        self.domains = OctObjectPool()
        self.nocts = 0 # Increment when initialized
        for i in range(3):
            self.DLE[i] = domain_left_edge[i] #0
            self.DRE[i] = domain_right_edge[i] #num_grid
        self._initialize_root_mesh()
        self.fill_style = "o"

    def _initialize_root_mesh(self):
        self.root_mesh = <Oct****> malloc(sizeof(void*) * self.nn[0])
        for i in range(self.nn[0]):
            self.root_mesh[i] = <Oct ***> malloc(sizeof(void*) * self.nn[1])
            for j in range(self.nn[1]):
                self.root_mesh[i][j] = <Oct **> malloc(sizeof(void*) * self.nn[2])
                for k in range(self.nn[2]):
                    self.root_mesh[i][j][k] = NULL

    @property
    def oct_arrays(self):
        return self.domains.to_arrays()

    @classmethod
    def load_octree(cls, header):
        cdef np.ndarray[np.uint8_t, ndim=1] ref_mask
        ref_mask = header['octree']
        cdef OctreeContainer obj = cls(header['dims'], header['left_edge'],
                header['right_edge'], num_zones = header['num_zones'],
                partial_coverage = header['partial_coverage'])
        # NOTE: We do not allow domain/file indices to be specified.
        cdef SelectorObject selector = AlwaysSelector(None)
        cdef oct_visitors.LoadOctree visitor
        visitor = oct_visitors.LoadOctree(obj, -1)
        cdef int i, j, k
        visitor.global_index = -1
        visitor.level = 0
        visitor.nz = visitor.nzones = 1
        assert(ref_mask.shape[0] / float(visitor.nzones) ==
            <int>(ref_mask.shape[0]/float(visitor.nzones)))
        obj.allocate_domains([ref_mask.shape[0] / visitor.nzones])
        cdef np.float64_t pos[3]
        cdef np.float64_t dds[3]
        # This dds is the oct-width
        for i in range(3):
            dds[i] = (obj.DRE[i] - obj.DLE[i]) / obj.nn[i]
        # Pos is the center of the octs
        cdef OctAllocationContainer *cur = obj.domains.get_cont(0)
        cdef Oct *o
        cdef np.uint64_t nfinest = 0
        visitor.ref_mask = ref_mask
        visitor.octs = cur.my_objs
        visitor.nocts = &cur.n_assigned
        visitor.nfinest = &nfinest
        pos[0] = obj.DLE[0] + dds[0]/2.0
        for i in range(obj.nn[0]):
            pos[1] = obj.DLE[1] + dds[1]/2.0
            for j in range(obj.nn[1]):
                pos[2] = obj.DLE[2] + dds[2]/2.0
                for k in range(obj.nn[2]):
                    if obj.root_mesh[i][j][k] != NULL:
                        raise RuntimeError
                    o = &cur.my_objs[cur.n_assigned]
                    o.domain_ind = o.file_ind = 0
                    o.domain = 1
                    obj.root_mesh[i][j][k] = o
                    cur.n_assigned += 1
                    visitor.pos[0] = i
                    visitor.pos[1] = j
                    visitor.pos[2] = k
                    # Always visit covered
                    selector.recursively_visit_octs(
                        obj.root_mesh[i][j][k],
                        pos, dds, 0, visitor, 1)
                    pos[2] += dds[2]
                pos[1] += dds[1]
            pos[0] += dds[0]
        obj.nocts = cur.n_assigned
        if obj.nocts * visitor.nz != ref_mask.size:
            raise KeyError(ref_mask.size, obj.nocts, obj.nz,
                obj.partial_coverage, visitor.nzones)
        return obj

    def __dealloc__(self):
        if self.root_mesh == NULL: return
        for i in range(self.nn[0]):
            if self.root_mesh[i] == NULL: continue
            for j in range(self.nn[1]):
                if self.root_mesh[i][j] == NULL: continue
                free(self.root_mesh[i][j])
            if self.root_mesh[i] == NULL: continue
            free(self.root_mesh[i])
        free(self.root_mesh)

    @cython.cdivision(True)
    cdef void visit_all_octs(self, SelectorObject selector,
                        OctVisitor visitor, int vc = -1,
                        np.int64_t *indices = NULL):
        cdef int i, j, k
        if vc == -1:
            vc = self.partial_coverage
        visitor.global_index = -1
        visitor.level = 0
        cdef np.float64_t pos[3]
        cdef np.float64_t dds[3]
        # This dds is the oct-width
        for i in range(3):
            dds[i] = (self.DRE[i] - self.DLE[i]) / self.nn[i]
        # Pos is the center of the octs
        pos[0] = self.DLE[0] + dds[0]/2.0
        for i in range(self.nn[0]):
            pos[1] = self.DLE[1] + dds[1]/2.0
            for j in range(self.nn[1]):
                pos[2] = self.DLE[2] + dds[2]/2.0
                for k in range(self.nn[2]):
                    if self.root_mesh[i][j][k] == NULL:
                        raise KeyError(i,j,k)
                    visitor.pos[0] = i
                    visitor.pos[1] = j
                    visitor.pos[2] = k
                    selector.recursively_visit_octs(
                        self.root_mesh[i][j][k],
                        pos, dds, 0, visitor, vc)
                    pos[2] += dds[2]
                pos[1] += dds[1]
            pos[0] += dds[0]

    cdef np.int64_t get_domain_offset(self, int domain_id):
        return 0

    def _check_root_mesh(self):
        cdef count = 0
        for i in range(self.nn[0]):
            for j in range(self.nn[1]):
                for k in range(self.nn[2]):
                    if self.root_mesh[i][j][k] == NULL:
                        print("Missing ", i, j, k)
                        count += 1
        print("Missing total of %s out of %s" % (count, self.nn[0] * self.nn[1] * self.nn[2]))

    cdef int get_root(self, int ind[3], Oct **o) nogil:
        cdef int i
        for i in range(3):
            if ind[i] < 0 or ind[i] >= self.nn[i]:
                o[0] = NULL
                return 1
        o[0] = self.root_mesh[ind[0]][ind[1]][ind[2]]
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef Oct *get(self, np.float64_t ppos[3], OctInfo *oinfo = NULL,
                  int max_level = 99) nogil:
        #Given a floating point position, retrieve the most
        #refined oct at that time
        cdef int ind32[3]
        cdef np.int64_t ipos[3]
        cdef np.float64_t dds[3]
        cdef np.float64_t cp[3]
        cdef Oct *cur
        cdef Oct *next
        cdef int i
        cur = next = NULL
        cdef np.int64_t ind[3]
        cdef np.int64_t level = -1
        for i in range(3):
            dds[i] = (self.DRE[i] - self.DLE[i])/self.nn[i]
            ind[i] = <np.int64_t> (floor((ppos[i] - self.DLE[i])/dds[i]))
            cp[i] = (ind[i] + 0.5) * dds[i] + self.DLE[i]
            ipos[i] = 0 # We add this to ind later, so it should be zero.
            ind32[i] = ind[i]
        self.get_root(ind32, &next)
        # We want to stop recursing when there's nowhere else to go
        while next != NULL and level < max_level:
            level += 1
            for i in range(3):
                ipos[i] = (ipos[i] << 1) + ind[i]
            cur = next
            for i in range(3):
                dds[i] = dds[i] / 2.0
                if cp[i] > ppos[i]:
                    ind[i] = 0
                    cp[i] -= dds[i] / 2.0
                else:
                    ind[i] = 1
                    cp[i] += dds[i]/2.0
            if cur.children != NULL:
                next = cur.children[cind(ind[0],ind[1],ind[2])]
            else:
                next = NULL
        if oinfo == NULL: return cur
        cdef np.float64_t factor = 1.0 / self.nz * 2
        for i in range(3):
            # We don't normally need to change dds[i] as it has been halved
            # from the oct width, thus making it already the cell width.
            # But, since not everything has the cell width equal to have the
            # width of the oct, we need to apply "factor".
            oinfo.dds[i] = dds[i] * factor # Cell width
            oinfo.ipos[i] = ipos[i]
            oinfo.left_edge[i] = oinfo.ipos[i] * (oinfo.dds[i] * self.nz) + self.DLE[i]
        oinfo.level = level
        return cur

    def locate_positions(self, np.float64_t[:,:] positions):
        """
        This routine, meant to be called by other internal routines, returns a
        list of oct IDs and a dictionary of Oct info for all the positions
        supplied.  Positions must be in code_length.
        """
        cdef np.float64_t factor = self.nz
        cdef dict all_octs = {}
        cdef OctInfo oi
        cdef Oct* o = NULL
        cdef np.float64_t pos[3]
        cdef np.ndarray[np.uint8_t, ndim=1] recorded
        cdef np.ndarray[np.int64_t, ndim=1] oct_id
        oct_id = np.ones(positions.shape[0], dtype="int64") * -1
        recorded = np.zeros(self.nocts, dtype="uint8")
        cdef np.int64_t i, j
        for i in range(positions.shape[0]):
            for j in range(3):
                pos[j] = positions[i,j]
            o = self.get(pos, &oi)
            if o == NULL:
                raise RuntimeError
            if recorded[o.domain_ind] == 0:
                left_edge = np.asarray(<np.float64_t[:3]>oi.left_edge).copy()
                dds = np.asarray(<np.float64_t[:3]>oi.dds).copy()
                right_edge = left_edge + dds*factor
                all_octs[o.domain_ind] = dict(
                    left_edge = left_edge,
                    right_edge = right_edge,
                    level = oi.level
                )
                recorded[o.domain_ind] = 1
            oct_id[i] = o.domain_ind
        return oct_id, all_octs

    def domain_identify(self, SelectorObject selector):
        cdef np.ndarray[np.uint8_t, ndim=1] domain_mask
        domain_mask = np.zeros(self.num_domains, dtype="uint8")
        cdef oct_visitors.IdentifyOcts visitor
        visitor = oct_visitors.IdentifyOcts(self)
        visitor.domain_mask = domain_mask
        self.visit_all_octs(selector, visitor)
        cdef int i
        domain_ids = []
        for i in range(self.num_domains):
            if domain_mask[i] == 1:
                domain_ids.append(i+1)
        return domain_ids

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef Oct** neighbors(self, OctInfo *oi, np.int64_t *nneighbors, Oct *o,
                         bint periodicity[3]):
        # We are going to do a brute-force search here.
        # This is not the most efficient -- in fact, it's relatively bad.  But
        # we will attempt to improve it in a future iteration, where we will
        # grow a stack of parent Octs.
        # Note that in the first iteration, we will just find the up-to-27
        # neighbors, including the main oct.
        cdef np.int64_t i, j, k, n, level, ii, dlevel
        cdef int ind[3]
        cdef OctList *olist
        cdef OctList *my_list
        my_list = olist = NULL
        cdef Oct *cand
        cdef np.int64_t npos[3]
        cdef np.int64_t ndim[3]
        # Now we get our boundaries for this level, so that we can wrap around
        # if need be.
        # ndim is the oct dimensions of the level, not the cell dimensions.
        for i in range(3):
            ndim[i] = <np.int64_t> ((self.DRE[i] - self.DLE[i]) / oi.dds[i])
            # Here we adjust for oi.dds meaning *cell* width.
            ndim[i] = (ndim[i] / self.nz)
        my_list = olist = OctList_append(NULL, o)
        for i in range(3):
            npos[0] = (oi.ipos[0] + (1 - i))
            if not periodicity[0] and not \
               (0 <= npos[0] < ndim[0]):
                continue
            elif npos[0] < 0: npos[0] += ndim[0]
            elif npos[0] >= ndim[0]: npos[0] -= ndim[0]
            for j in range(3):
                npos[1] = (oi.ipos[1] + (1 - j))
                if not periodicity[1] and not \
                   (0 <= npos[1] < ndim[1]):
                    continue
                elif npos[1] < 0: npos[1] += ndim[1]
                elif npos[1] >= ndim[1]: npos[1] -= ndim[1]
                for k in range(3):
                    npos[2] = (oi.ipos[2] + (1 - k))
                    if not periodicity[2] and not \
                       (0 <= npos[2] < ndim[2]):
                        continue
                    if npos[2] < 0: npos[2] += ndim[2]
                    if npos[2] >= ndim[2]: npos[2] -= ndim[2]
                    # Now we have our npos, which we just need to find.
                    # Level 0 gets bootstrapped
                    for n in range(3):
                        ind[n] = ((npos[n] >> (oi.level)) & 1)
                    cand = NULL
                    self.get_root(ind, &cand)
                    # We should not get a NULL if we handle periodicity
                    # correctly, but we might.
                    if cand == NULL: continue
                    for level in range(1, oi.level+1):
                        dlevel = oi.level - level
                        if cand.children == NULL: break
                        for n in range(3):
                            ind[n] = (npos[n] >> dlevel) & 1
                        ii = cind(ind[0],ind[1],ind[2])
                        if cand.children[ii] == NULL: break
                        cand = cand.children[ii]
                    if cand.children != NULL:
                        olist = OctList_subneighbor_find(
                            olist, cand, i, j, k)
                    else:
                        olist = OctList_append(olist, cand)
        olist = my_list
        cdef int noct = OctList_count(olist)
        cdef Oct **neighbors
        neighbors = <Oct **> malloc(sizeof(Oct*)*noct)
        for i in range(noct):
            neighbors[i] = olist.o
            olist = olist.next
        OctList_delete(my_list)
        nneighbors[0] = noct
        return neighbors

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def mask(self, SelectorObject selector, np.int64_t num_cells = -1,
             int domain_id = -1):
        if num_cells == -1:
            num_cells = selector.count_octs(self, domain_id)
        cdef np.ndarray[np.uint8_t, ndim=4] mask
        cdef oct_visitors.MaskOcts visitor
        visitor = oct_visitors.MaskOcts(self, domain_id)
        cdef int ns = self.nz
        mask = np.zeros((num_cells, ns, ns, ns), dtype="uint8")
        visitor.mask = mask
        self.visit_all_octs(selector, visitor)
        return mask.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def icoords(self, SelectorObject selector, np.int64_t num_cells = -1,
                int domain_id = -1):
        if num_cells == -1:
            num_cells = selector.count_oct_cells(self, domain_id)
        cdef oct_visitors.ICoordsOcts visitor
        visitor = oct_visitors.ICoordsOcts(self, domain_id)
        cdef np.ndarray[np.int64_t, ndim=2] coords
        coords = np.empty((num_cells, 3), dtype="int64")
        visitor.icoords = coords
        self.visit_all_octs(selector, visitor)
        return coords

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def ires(self, SelectorObject selector, np.int64_t num_cells = -1,
                int domain_id = -1):
        cdef int i
        if num_cells == -1:
            num_cells = selector.count_oct_cells(self, domain_id)
        cdef oct_visitors.IResOcts visitor
        visitor = oct_visitors.IResOcts(self, domain_id)
        #Return the 'resolution' of each cell; ie the level
        cdef np.ndarray[np.int64_t, ndim=1] res
        res = np.empty(num_cells, dtype="int64")
        visitor.ires = res
        self.visit_all_octs(selector, visitor)
        if self.level_offset > 0:
            for i in range(num_cells):
                res[i] += self.level_offset
        return res

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fwidth(self, SelectorObject selector, np.int64_t num_cells = -1,
                int domain_id = -1):
        if num_cells == -1:
            num_cells = selector.count_oct_cells(self, domain_id)
        cdef oct_visitors.FWidthOcts visitor
        visitor = oct_visitors.FWidthOcts(self, domain_id)
        cdef np.ndarray[np.float64_t, ndim=2] fwidth
        fwidth = np.empty((num_cells, 3), dtype="float64")
        visitor.fwidth = fwidth
        self.visit_all_octs(selector, visitor)
        cdef np.float64_t base_dx
        for i in range(3):
            base_dx = (self.DRE[i] - self.DLE[i])/self.nn[i]
            fwidth[:,i] *= base_dx
        return fwidth

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fcoords(self, SelectorObject selector, np.int64_t num_cells = -1,
                int domain_id = -1):
        if num_cells == -1:
            num_cells = selector.count_oct_cells(self, domain_id)
        cdef oct_visitors.FCoordsOcts visitor
        visitor = oct_visitors.FCoordsOcts(self, domain_id)
        #Return the floating point unitary position of every cell
        cdef np.ndarray[np.float64_t, ndim=2] coords
        coords = np.empty((num_cells, 3), dtype="float64")
        visitor.fcoords = coords
        self.visit_all_octs(selector, visitor)
        cdef int i
        cdef np.float64_t base_dx
        for i in range(3):
            base_dx = (self.DRE[i] - self.DLE[i])/self.nn[i]
            coords[:,i] *= base_dx
            coords[:,i] += self.DLE[i]
        return coords

    def save_octree(self):
        # Get the header
        header = dict(dims = (self.nn[0], self.nn[1], self.nn[2]),
                      left_edge = (self.DLE[0], self.DLE[1], self.DLE[2]),
                      right_edge = (self.DRE[0], self.DRE[1], self.DRE[2]),
                      num_zones = self.nz,
                      partial_coverage = self.partial_coverage)
        cdef SelectorObject selector = AlwaysSelector(None)
        # domain_id = -1 here, because we want *every* oct
        cdef oct_visitors.StoreOctree visitor
        visitor = oct_visitors.StoreOctree(self, -1)
        visitor.nz = 1
        cdef np.ndarray[np.uint8_t, ndim=1] ref_mask
        ref_mask = np.zeros(self.nocts * visitor.nzones, dtype="uint8") - 1
        visitor.ref_mask = ref_mask
        # Enforce partial_coverage here
        self.visit_all_octs(selector, visitor, 1)
        header['octree'] = ref_mask
        return header

    def selector_fill(self, SelectorObject selector,
                      np.ndarray source,
                      np.ndarray dest = None,
                      np.int64_t offset = 0, int dims = 1,
                      int domain_id = -1):
        # This is actually not correct.  The hard part is that we need to
        # iterate the same way visit_all_octs does, but we need to track the
        # number of octs total visited.
        cdef np.int64_t num_cells = -1
        if dest is None:
            # Note that RAMSES can have partial refinement inside an Oct.  This
            # means we actually do want the number of Octs, not the number of
            # cells.
            num_cells = selector.count_oct_cells(self, domain_id)
            dest = np.zeros((num_cells, dims), dtype=source.dtype,
                            order='C')
        if dims != 1:
            raise RuntimeError
        # Just make sure that we're in the right shape.  Ideally this will not
        # duplicate memory.  Since we're in Cython, we want to avoid modifying
        # the .shape attributes directly.
        dest = dest.reshape((num_cells, 1))
        source = source.reshape((source.shape[0], source.shape[1],
                    source.shape[2], source.shape[3], dims))
        cdef OctVisitor visitor
        cdef oct_visitors.CopyArrayI64 visitor_i64
        cdef oct_visitors.CopyArrayF64 visitor_f64
        if source.dtype != dest.dtype:
            raise RuntimeError
        if source.dtype == np.int64:
            visitor_i64 = oct_visitors.CopyArrayI64(self, domain_id)
            visitor_i64.source = source
            visitor_i64.dest = dest
            visitor = visitor_i64
        elif source.dtype == np.float64:
            visitor_f64 = oct_visitors.CopyArrayF64(self, domain_id)
            visitor_f64.source = source
            visitor_f64.dest = dest
            visitor = visitor_f64
        else:
            raise NotImplementedError
        visitor.index = offset
        # We only need this so we can continue calculating the offset
        visitor.dims = dims
        self.visit_all_octs(selector, visitor)
        if (visitor.global_index + 1) * visitor.nzones * visitor.dims > source.size:
            print("GLOBAL INDEX RAN AHEAD.",)
            print (visitor.global_index + 1) * visitor.nzones * visitor.dims - source.size
            print(dest.size, source.size, num_cells)
            raise RuntimeError
        if visitor.index > dest.size:
            print("DEST INDEX RAN AHEAD.",)
            print(visitor.index - dest.size)
            print (visitor.global_index + 1) * visitor.nzones * visitor.dims, source.size
            print(num_cells)
            raise RuntimeError
        if num_cells >= 0:
            return dest
        return visitor.index - offset

    def domain_ind(self, selector, int domain_id = -1):
        cdef np.ndarray[np.int64_t, ndim=1] ind
        # Here's where we grab the masked items.
        ind = np.full(self.nocts, -1, 'int64')
        cdef oct_visitors.IndexOcts visitor
        visitor = oct_visitors.IndexOcts(self, domain_id)
        visitor.oct_index = ind
        self.visit_all_octs(selector, visitor)
        return ind

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def add(self, int curdom, int curlevel,
            np.ndarray[np.float64_t, ndim=2] pos,
            int skip_boundary = 1,
            int count_boundary = 0,
            np.ndarray[np.uint64_t, ndim=1, cast=True] levels = None
            ):
        # In this function, if we specify curlevel = -1, then we query the
        # (optional) levels array for the oct level.
        cdef int no, p, i
        cdef int ind[3]
        cdef int nb = 0
        cdef Oct *cur
        cdef np.float64_t pp[3]
        cdef np.float64_t cp[3]
        cdef np.float64_t dds[3]
        no = pos.shape[0] #number of octs
        if curdom > self.num_domains: return 0
        cdef OctAllocationContainer *cont = self.domains.get_cont(curdom - 1)
        cdef int initial = cont.n_assigned
        cdef int in_boundary = 0
        # How do we bootstrap ourselves?
        for p in range(no):
            # We allow specifying curlevel = -1 to query from the levels array
            # instead.
            if curlevel == -1:
                this_level = levels[p]
            else:
                this_level = curlevel
            #for every oct we're trying to add find the
            #floating point unitary position on this level
            in_boundary = 0
            for i in range(3):
                pp[i] = pos[p, i]
                dds[i] = (self.DRE[i] - self.DLE[i])/self.nn[i]
                ind[i] = <np.int64_t> ((pp[i] - self.DLE[i])/dds[i])
                cp[i] = (ind[i] + 0.5) * dds[i] + self.DLE[i]
                if ind[i] < 0 or ind[i] >= self.nn[i]:
                    in_boundary = 1
            if skip_boundary == in_boundary == 1:
                nb += count_boundary
                continue
            cur = self.next_root(curdom, ind)
            if cur == NULL: raise RuntimeError
            # Now we find the location we want
            # Note that RAMSES I think 1-findiceses levels, but we don't.
            for _ in range(this_level):
                # At every level, find the cell this oct
                # lives inside
                for i in range(3):
                    #as we get deeper, oct size halves
                    dds[i] = dds[i] / 2.0
                    if cp[i] > pp[i]:
                        ind[i] = 0
                        cp[i] -= dds[i]/2.0
                    else:
                        ind[i] = 1
                        cp[i] += dds[i]/2.0
                # Check if it has not been allocated
                cur = self.next_child(curdom, ind, cur)
            # Now we should be at the right level
            cur.domain = curdom
            cur.file_ind = p
        return cont.n_assigned - initial + nb

    def allocate_domains(self, domain_counts):
        cdef int count, i
        self.num_domains = len(domain_counts) # 1-indexed
        for i, count in enumerate(domain_counts):
            self.domains.append(count)

    cdef void append_domain(self, np.int64_t domain_count):
        self.num_domains += 1
        self.domains.append(domain_count)

    cdef Oct* next_root(self, int domain_id, int ind[3]):
        cdef Oct *next = self.root_mesh[ind[0]][ind[1]][ind[2]]
        if next != NULL: return next
        cdef OctAllocationContainer *cont = self.domains.get_cont(domain_id - 1)
        if cont.n_assigned >= cont.n: raise RuntimeError
        next = &cont.my_objs[cont.n_assigned]
        cont.n_assigned += 1
        self.root_mesh[ind[0]][ind[1]][ind[2]] = next
        self.nocts += 1
        return next

    cdef Oct* next_child(self, int domain_id, int ind[3], Oct *parent) except? NULL:
        cdef int i
        cdef Oct *next = NULL
        if parent.children != NULL:
            next = parent.children[cind(ind[0],ind[1],ind[2])]
        else:
            # This *8 does NOT need to be made generic.
            parent.children = <Oct **> malloc(sizeof(Oct *) * 8)
            for i in range(8):
                parent.children[i] = NULL
        if next != NULL: return next
        cdef OctAllocationContainer *cont = self.domains.get_cont(domain_id - 1)
        if cont.n_assigned >= cont.n: raise RuntimeError
        next = &cont.my_objs[cont.n_assigned]
        cont.n_assigned += 1
        parent.children[cind(ind[0],ind[1],ind[2])] = next
        self.nocts += 1
        return next

    def file_index_octs(self, SelectorObject selector, int domain_id,
                        num_cells = -1):
        # We create oct arrays of the correct size
        cdef np.ndarray[np.uint8_t, ndim=1] levels
        cdef np.ndarray[np.uint8_t, ndim=1] cell_inds
        cdef np.ndarray[np.int64_t, ndim=1] file_inds
        if num_cells < 0:
            num_cells = selector.count_oct_cells(self, domain_id)
        # Initialize variables with dummy values
        levels = np.full(num_cells, 255, dtype="uint8")
        file_inds = np.full(num_cells, -1, dtype="int64")
        cell_inds = np.full(num_cells, 8, dtype="uint8")
        cdef oct_visitors.FillFileIndicesO visitor_o
        cdef oct_visitors.FillFileIndicesR visitor_r
        if self.fill_style == "r":
            visitor_r = oct_visitors.FillFileIndicesR(self, domain_id)
            visitor_r.levels = levels
            visitor_r.file_inds = file_inds
            visitor_r.cell_inds = cell_inds
            visitor = visitor_r
        elif self.fill_style == "o":
            visitor_o = oct_visitors.FillFileIndicesO(self, domain_id)
            visitor_o.levels = levels
            visitor_o.file_inds = file_inds
            visitor_o.cell_inds = cell_inds
            visitor = visitor_o
        else:
            raise RuntimeError
        self.visit_all_octs(selector, visitor)
        return levels, cell_inds, file_inds

    def morton_index_octs(self, SelectorObject selector, int domain_id,
                          num_cells = -1):
        cdef np.int64_t i
        cdef np.uint8_t[:] levels
        cdef np.uint64_t[:] morton_inds
        if num_cells < 0:
            num_cells = selector.count_oct_cells(self, domain_id)
        levels = np.zeros(num_cells, dtype="uint8")
        morton_inds = np.zeros(num_cells, dtype="uint64")
        for i in range(num_cells):
            levels[i] = 100
            morton_inds[i] = 0
        cdef oct_visitors.MortonIndexOcts visitor
        visitor = oct_visitors.MortonIndexOcts(self, domain_id)
        visitor.level_arr = levels
        visitor.morton_ind = morton_inds
        self.visit_all_octs(selector, visitor)
        return levels, morton_inds

    def domain_count(self, SelectorObject selector):
        # We create oct arrays of the correct size
        cdef np.ndarray[np.int64_t, ndim=1] domain_counts
        domain_counts = np.zeros(self.num_domains, dtype="int64")
        cdef oct_visitors.CountByDomain visitor
        visitor = oct_visitors.CountByDomain(self, -1)
        visitor.domain_counts = domain_counts
        self.visit_all_octs(selector, visitor)
        return domain_counts

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_level(self, int level,
                   np.ndarray[np.uint8_t, ndim=1] levels,
                   np.ndarray[np.uint8_t, ndim=1] cell_inds,
                   np.ndarray[np.int64_t, ndim=1] file_inds,
                   dest_fields, source_fields,
                   np.int64_t offset = 0):
        cdef np.ndarray[np.float64_t, ndim=2] source
        cdef np.ndarray[np.float64_t, ndim=1] dest
        cdef int i, lvl

        for key in dest_fields:
            dest = dest_fields[key]
            source = source_fields[key]
            for i in range(levels.shape[0]):
                lvl = levels[i]
                if lvl != level: continue
                if file_inds[i] < 0:
                    dest[i + offset] = np.nan
                else:
                    dest[i + offset] = source[file_inds[i], cell_inds[i]]

    def fill_index(self, SelectorObject selector = AlwaysSelector(None)):
        """Get the on-file index of each cell"""
        cdef StoreIndex visitor

        cdef np.int64_t[:, :, :, :] cell_inds

        cell_inds = np.full((self.nocts, 2, 2, 2), -1, dtype=np.int64)

        visitor = StoreIndex(self, -1)
        visitor.cell_inds = cell_inds

        self.visit_all_octs(selector, visitor)

        return np.asarray(cell_inds)

    def fill_octcellindex_neighbours(self, SelectorObject selector, int num_octs=-1, int domain_id=-1, int n_ghost_zones=1):
        """Compute the oct and cell indices of all the cells within all selected octs, extended
        by one cell in all directions (for ghost zones computations).

        Parameters
        ----------
        selector : SelectorObject
            Selector for the octs to compute neighbour of
        num_octs : int, optional
            The number of octs to read in
        domain_id : int, optional
            The domain to perform the selection over

        Returns
        -------
        oct_inds : int64 ndarray (nocts*8, )
            The on-domain index of the octs containing each cell
        cell_inds : uint8 ndarray (nocts*8, )
            The index of the cell in its parent oct

        Note
        ----
        oct_inds/cell_inds
        """
        if num_octs == -1:
            num_octs = selector.count_octs(self, domain_id)

        cdef NeighbourCellIndexVisitor visitor

        cdef np.uint8_t[::1] cell_inds
        cdef np.int64_t[::1] oct_inds

        cell_inds = np.full(num_octs*4**3, 8, dtype=np.uint8)
        oct_inds = np.full(num_octs*4**3, -1, dtype=np.int64)

        visitor = NeighbourCellIndexVisitor(self, -1, n_ghost_zones)
        visitor.cell_inds = cell_inds
        visitor.domain_inds = oct_inds

        self.visit_all_octs(selector, visitor)

        return np.asarray(oct_inds), np.asarray(cell_inds)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_level_with_domain(
                   self, int level,
                   np.uint8_t[:] levels,
                   np.uint8_t[:] cell_inds,
                   np.int64_t[:] file_inds,
                   np.int32_t[:] domains,
                   dict dest_fields,
                   dict source_fields,
                   np.int32_t domain,
                   np.int64_t offset = 0
                   ):
        """Similar to fill_level but accepts a domain argument.

        This is particularly useful for frontends that have buffer zones at CPU boundaries.
        These buffer oct cells have a different domain than the local one and
        are usually not read, but one has to read them e.g. to compute ghost zones.
        """
        cdef np.ndarray[np.float64_t, ndim=2] source
        cdef np.ndarray[np.float64_t, ndim=1] dest
        cdef int i, count, lev
        cdef np.int32_t dom

        for key in dest_fields:
            dest = dest_fields[key]
            source = source_fields[key]
            count = 0
            for i in range(levels.shape[0]):
                lev = levels[i]
                dom = domains[i]
                if lev != level or dom != domain: continue
                count += 1
                if file_inds[i] < 0:
                    dest[i + offset] = np.nan
                else:
                    dest[i + offset] = source[file_inds[i], cell_inds[i]]
        return count

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def file_index_octs_with_ghost_zones(
            self, SelectorObject selector, int domain_id,
            int num_cells=1, int n_ghost_zones=1):
        """Similar as file_index_octs, but returns the level, cell index,
        file index and domain of the neighbouring cells as well.

        Arguments
        ---------
        selector : SelectorObject
            The selector object. It is expected to select all cells for a given oct.
        domain_id : int
            The domain to select. Set to -1 to select all domains.
        num_cells : int, optional
            The total number of cells (accounting for a 1-cell thick ghost zone layer).

        Returns
        -------
        levels : uint8, shape (num_cells,)
            The level of each cell of the super oct
        cell_inds : uint8, shape (num_cells, )
            The index of each cell of the super oct within its own oct
        file_inds : int64, shape (num_cells, )
            The on-file position of the cell. See notes below.
        domains : int32, shape (num_cells)
            The domain to which the cells belongs. See notes below.

        Notes
        -----

        The algorithm constructs a "super-oct" around each oct (see sketch below,
        where the original oct cells are marked with an x).

        Note that for sparse octrees (such as RAMSES'), the neighbouring cells
        may belong to another domain (this is stored in `domains`). If the dataset
        provides buffer zones between domains (such as RAMSES), this may be stored
        locally and can be accessed directly.


        +---+---+---+---+
        |   |   |   |   |
        |---+---+---+---|
        |   | x | x |   |
        |---+---+---+---|
        |   | x | x |   |
        |---+---+---+---|
        |   |   |   |   |
        +---+---+---+---+

        """
        cdef int num_octs
        if num_cells < 0:
            num_octs = selector.count_octs(self, domain_id)
            num_cells = num_octs * 4**3
        cdef NeighbourCellVisitor visitor

        cdef np.ndarray[np.uint8_t, ndim=1] levels
        cdef np.ndarray[np.uint8_t, ndim=1] cell_inds
        cdef np.ndarray[np.int64_t, ndim=1] file_inds
        cdef np.ndarray[np.int32_t, ndim=1] domains
        levels = np.full(num_cells, 255, dtype="uint8")
        file_inds = np.full(num_cells, -1, dtype="int64")
        cell_inds = np.full(num_cells, 8, dtype="uint8")
        domains = np.full(num_cells, -1, dtype="int32")

        visitor = NeighbourCellVisitor(self, -1, n_ghost_zones)
        # output: level, file_ind and cell_ind of the neighbouring cells
        visitor.levels = levels
        visitor.file_inds = file_inds
        visitor.cell_inds = cell_inds
        visitor.domains = domains
        # direction to explore and extra parameters of the visitor
        visitor.octree = self
        visitor.last = -1

        # Compute indices
        self.visit_all_octs(selector, visitor)

        return levels, cell_inds, file_inds, domains

    def finalize(self):
        cdef SelectorObject selector = AlwaysSelector(None)
        cdef oct_visitors.AssignDomainInd visitor
        visitor = oct_visitors.AssignDomainInd(self, 1)
        self.visit_all_octs(selector, visitor)
        assert ((visitor.global_index+1)*visitor.nzones == visitor.index)

cdef int root_node_compare(const void *a, const void *b) noexcept nogil:
    cdef OctKey *ao
    cdef OctKey *bo
    ao = <OctKey *>a
    bo = <OctKey *>b
    if ao.key < bo.key:
        return -1
    elif ao.key == bo.key:
        return 0
    else:
        return 1

cdef class SparseOctreeContainer(OctreeContainer):

    def __init__(self, domain_dimensions, domain_left_edge, domain_right_edge,
                 num_zones = 2):
        cdef int i
        self.partial_coverage = 1
        self.nz = num_zones
        for i in range(3):
            self.nn[i] = domain_dimensions[i]
        self.domains = OctObjectPool()
        self.num_domains = 0
        self.level_offset = 0
        self.nocts = 0 # Increment when initialized
        self.root_mesh = NULL
        self.root_nodes = NULL
        self.tree_root = NULL
        self.num_root = 0
        self.max_root = 0
        # We don't initialize the octs yet
        for i in range(3):
            self.DLE[i] = domain_left_edge[i] #0
            self.DRE[i] = domain_right_edge[i] #num_grid
        self.fill_style = "r"

    @classmethod
    def load_octree(cls, header):
        raise NotImplementedError

    def save_octree(self):
        raise NotImplementedError

    cdef int get_root(self, int ind[3], Oct **o) nogil:
        o[0] = NULL
        cdef np.int64_t key = self.ipos_to_key(ind)
        cdef OctKey okey
        cdef OctKey **oresult = NULL
        okey.key = key
        okey.node = NULL
        oresult = <OctKey **> tfind(<void*>&okey,
            &self.tree_root, root_node_compare)
        if oresult != NULL:
            o[0] = oresult[0].node
            return 1
        return 0

    cdef void key_to_ipos(self, np.int64_t key, np.int64_t pos[3]):
        # Note: this is the result of doing
        # for i in range(20):
        #     ukey |= (1 << i)
        cdef np.int64_t ukey = 1048575
        cdef int j
        for j in range(3):
            pos[2 - j] = (<np.int64_t>(key & ukey))
            key = key >> 20

    cdef np.int64_t ipos_to_key(self, int pos[3]) nogil:
        # We (hope) that 20 bits is enough for each index.
        cdef int i
        cdef np.int64_t key = 0
        for i in range(3):
            # Note the casting here.  Bitshifting can cause issues otherwise.
            key |= ((<np.int64_t>pos[i]) << 20 * (2 - i))
        return key

    @cython.cdivision(True)
    cdef void visit_all_octs(self, SelectorObject selector,
                        OctVisitor visitor,
                        int vc = -1,
                        np.int64_t *indices = NULL):
        cdef int i, j
        cdef np.int64_t key
        visitor.global_index = -1
        visitor.level = 0
        if vc == -1:
            vc = self.partial_coverage
        cdef np.float64_t pos[3]
        cdef np.float64_t dds[3]
        # This dds is the oct-width
        for i in range(3):
            dds[i] = (self.DRE[i] - self.DLE[i]) / self.nn[i]
        # Pos is the center of the octs
        cdef Oct *o
        for i in range(self.num_root):
            o = self.root_nodes[i].node
            key = self.root_nodes[i].key
            self.key_to_ipos(key, visitor.pos)
            for j in range(3):
                pos[j] = self.DLE[j] + (visitor.pos[j] + 0.5) * dds[j]
            selector.recursively_visit_octs(
                o, pos, dds, 0, visitor, vc)
            if indices != NULL:
                indices[i] = visitor.index

    cdef np.int64_t get_domain_offset(self, int domain_id):
        return 0 # We no longer have a domain offset.

    cdef Oct* next_root(self, int domain_id, int ind[3]):
        cdef Oct *next = NULL
        self.get_root(ind, &next)
        if next != NULL: return next
        cdef OctAllocationContainer *cont = self.domains.get_cont(domain_id - 1)
        if cont.n_assigned >= cont.n:
            print("Too many assigned.")
            return NULL
        if self.num_root >= self.max_root:
            print("Too many roots.")
            return NULL
        next = &cont.my_objs[cont.n_assigned]
        cont.n_assigned += 1
        cdef np.int64_t key = 0
        cdef OctKey *ikey = &self.root_nodes[self.num_root]
        key = self.ipos_to_key(ind)
        self.root_nodes[self.num_root].key = key
        self.root_nodes[self.num_root].node = next
        tsearch(<void*>ikey, &self.tree_root, root_node_compare)
        self.num_root += 1
        self.nocts += 1
        return next

    def allocate_domains(self, domain_counts, int root_nodes):
        OctreeContainer.allocate_domains(self, domain_counts)
        self.root_nodes = <OctKey*> malloc(sizeof(OctKey) * root_nodes)
        self.max_root = root_nodes
        for i in range(root_nodes):
            self.root_nodes[i].key = -1
            self.root_nodes[i].node = NULL

    def __dealloc__(self):
        # This gets called BEFORE the superclass deallocation.  But, both get
        # called.
        cdef OctKey *ikey
        for i in range(self.num_root):
            ikey = &self.root_nodes[i]
            tdelete(<void *>ikey, &self.tree_root, root_node_compare)

        if self.root_nodes != NULL: free(self.root_nodes)

cdef class ARTOctreeContainer(OctreeContainer):
    def __init__(self, oct_domain_dimensions, domain_left_edge,
                 domain_right_edge, partial_coverage = 0,
                 num_zones = 2):
        OctreeContainer.__init__(self, oct_domain_dimensions,
                domain_left_edge, domain_right_edge, partial_coverage,
                 num_zones)
        self.fill_style = "r"

cdef OctList *OctList_subneighbor_find(OctList *olist, Oct *top,
                                       int i, int j, int k):
    if top.children == NULL: return olist
    # The i, j, k here are the offsets of "top" with respect to
    # the oct for whose neighbors we are searching.
    # Note that this will be recursively called.  We will evaluate either 1, 2,
    # or 4 octs for children and potentially adding them.  In fact, this will
    # be 2**(num_zero) where num_zero is the number of indices that are equal
    # to zero; i.e., the number of dimensions along which we are aligned.
    # For now, we assume we will not be doing this along all three zeros,
    # because that would be pretty tricky.
    if i == j == k == 1: return olist
    cdef np.int64_t n[3]
    cdef np.int64_t ind[3]
    cdef np.int64_t off[3][2]
    cdef np.int64_t ii, ij, ik, ci
    ind[0] = 1 - i
    ind[1] = 1 - j
    ind[2] = 1 - k
    for ii in range(3):
        if ind[ii] == 0:
            n[ii] = 2
            off[ii][0] = 0
            off[ii][1] = 1
        elif ind[ii] == -1:
            n[ii] = 1
            off[ii][0] = 1
        elif ind[ii] == 1:
            n[ii] = 1
            off[ii][0] = 0
    for ii in range(n[0]):
        for ij in range(n[1]):
            for ik in range(n[2]):
                ci = cind(off[0][ii], off[1][ij], off[2][ik])
                cand = top.children[ci]
                if cand.children != NULL:
                    olist = OctList_subneighbor_find(olist,
                        cand, i, j, k)
                else:
                    olist = OctList_append(olist, cand)
    return olist

cdef OctList *OctList_append(OctList *olist, Oct *o):
    cdef OctList *this = olist
    if this == NULL:
        this = <OctList *> malloc(sizeof(OctList))
        this.next = NULL
        this.o = o
        return this
    while this.next != NULL:
        this = this.next
    this.next = <OctList*> malloc(sizeof(OctList))
    this = this.next
    this.o = o
    this.next = NULL
    return this

cdef int OctList_count(OctList *olist):
    cdef OctList *this = olist
    cdef int i = 0 # Count the list
    while this != NULL:
        i += 1
        this = this.next
    return i

cdef void OctList_delete(OctList *olist):
    cdef OctList *next
    cdef OctList *this = olist
    while this != NULL:
        next = this.next
        free(this)
        this = next

cdef class OctObjectPool(ObjectPool):
    # This is an inherited version of the ObjectPool that provides setup and
    # teardown functions for the individually allocated objects.  These allow
    # us to initialize the Octs to default values, and we can also free any
    # allocated memory in them.  Implementing _con_to_array also provides the
    # opportunity to supply views of the octs in Python code.
    def __cinit__(self):
        # Base class will ALSO be called
        self.itemsize = sizeof(Oct)
        assert(sizeof(OctAllocationContainer) == sizeof(AllocationContainer))

    cdef void setup_objs(self, void *obj, np.uint64_t n, np.uint64_t offset,
                         np.int64_t con_id):
        cdef Oct* octs = <Oct *> obj
        for n in range(n):
            octs[n].file_ind = octs[n].domain = - 1
            octs[n].domain_ind = n + offset
            octs[n].children = NULL

    cdef void teardown_objs(self, void *obj, np.uint64_t n, np.uint64_t offset,
                           np.int64_t con_id):
        cdef np.uint64_t i
        cdef Oct *my_octs = <Oct *> obj
        for i in range(n):
            if my_octs[i].children != NULL:
                free(my_octs[i].children)
        free(obj)

    def _con_to_array(self, int i):
        cdef AllocationContainer *obj = &self.containers[i]
        if obj.n_assigned == 0:
            return None
        cdef OctPadded[:] mm = <OctPadded[:obj.n_assigned]> (
                <OctPadded*> obj.my_objs)
        rv = np.asarray(mm)
        return rv



cdef class CuttingPlaneSelector(SelectorObject):
    cdef public np.float64_t norm_vec[3]
    cdef public np.float64_t d

    def __init__(self, dobj):
        cdef int i
        for i in range(3):
            self.norm_vec[i] = dobj._norm_vec[i]
        self.d = _ensure_code(dobj._d)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        cdef np.float64_t left_edge[3]
        cdef np.float64_t right_edge[3]
        cdef int i
        for i in range(3):
            left_edge[i] = pos[i] - 0.5*dds[i]
            right_edge[i] = pos[i] + 0.5*dds[i]
        return self.select_bbox(left_edge, right_edge)

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        # two 0-volume constructs don't intersect
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        cdef int i
        cdef np.float64_t height = self.d
        for i in range(3) :
            height += pos[i] * self.norm_vec[i]
        if height*height <= radius*radius : return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        cdef int i, j, k, n
        cdef np.float64_t *arr[2]
        cdef np.float64_t pos[3]
        cdef np.float64_t gd
        arr[0] = left_edge
        arr[1] = right_edge
        all_under = 1
        all_over = 1
        # Check each corner
        for i in range(2):
            pos[0] = arr[i][0]
            for j in range(2):
                pos[1] = arr[j][1]
                for k in range(2):
                    pos[2] = arr[k][2]
                    gd = self.d
                    for n in range(3):
                        gd += pos[n] * self.norm_vec[n]
                    # this allows corners and faces on the low-end to
                    # collide, while not selecting cells on the high-side
                    if i == 0 and j == 0 and k == 0 :
                        if gd <= 0: all_over = 0
                        if gd >= 0: all_under = 0
                    else :
                        if gd < 0: all_over = 0
                        if gd > 0: all_under = 0
        if all_over == 1 or all_under == 1:
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        cdef int i, j, k, n
        cdef np.float64_t *arr[2]
        cdef np.float64_t pos[3]
        cdef np.float64_t gd
        arr[0] = left_edge
        arr[1] = right_edge
        all_under = 1
        all_over = 1
        # Check each corner
        for i in range(2):
            pos[0] = arr[i][0]
            for j in range(2):
                pos[1] = arr[j][1]
                for k in range(2):
                    pos[2] = arr[k][2]
                    gd = self.d
                    for n in range(3):
                        gd += pos[n] * self.norm_vec[n]
                    # this allows corners and faces on the low-end to
                    # collide, while not selecting cells on the high-side
                    if i == 0 and j == 0 and k == 0 :
                        if gd <= 0: all_over = 0
                        if gd >= 0: all_under = 0
                    else :
                        if gd < 0: all_over = 0
                        if gd > 0: all_under = 0
        if all_over == 1 or all_under == 1:
            return 0
        return 2 # a box of non-zeros volume can't be inside a plane

    def _hash_vals(self):
        return (("norm_vec[0]", self.norm_vec[0]),
                ("norm_vec[1]", self.norm_vec[1]),
                ("norm_vec[2]", self.norm_vec[2]),
                ("d", self.d))

    def _get_state_attnames(self):
        return ("d", "norm_vec")

cutting_selector = CuttingPlaneSelector



import sys
from collections import UserDict
from collections.abc import Callable
from typing import Dict, List, Optional, Tuple

from unyt.exceptions import UnitConversionError

from yt._maintenance.deprecation import issue_deprecation_warning
from yt._typing import FieldKey, FieldName, FieldType, KnownFieldsT
from yt.config import ytcfg
from yt.fields.field_exceptions import NeedsConfiguration
from yt.funcs import mylog, obj_length, only_on_root
from yt.geometry.api import Geometry
from yt.units.dimensions import dimensionless  # type: ignore
from yt.units.unit_object import Unit  # type: ignore
from yt.utilities.exceptions import (
    YTCoordinateNotImplemented,
    YTDomainOverflow,
    YTFieldNotFound,
)

from .derived_field import DeprecatedFieldFunc, DerivedField, NullFunc, TranslationFunc
from .field_plugin_registry import FunctionName, field_plugins
from .particle_fields import (
    add_union_field,
    particle_deposition_functions,
    particle_scalar_functions,
    particle_vector_functions,
    sph_whitelist_fields,
    standard_particle_fields,
)

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never


class FieldInfoContainer(UserDict):
    """
    This is a generic field container.  It contains a list of potential derived
    fields, all of which know how to act on a data object and return a value.
    This object handles converting units as well as validating the availability
    of a given field.

    """

    fallback = None
    known_other_fields: KnownFieldsT = ()
    known_particle_fields: KnownFieldsT = ()
    extra_union_fields: Tuple[FieldKey, ...] = ()

    def __init__(self, ds, field_list: List[FieldKey], slice_info=None):
        super().__init__()
        self._show_field_errors: List[Exception] = []
        self.ds = ds
        # Now we start setting things up.
        self.field_list = field_list
        self.slice_info = slice_info
        self.field_aliases: Dict[FieldKey, FieldKey] = {}
        self.species_names: List[FieldName] = []
        self.setup_fluid_aliases()

    @property
    def curvilinear(self) -> bool:
        issue_deprecation_warning(
            "FieldInfoContainer.curvilinear attribute is deprecated. "
            "Please compare the internal dataset geometry directly to known Geometry enum members instead. ",
            since="4.2",
        )
        geometry = self.ds.geometry
        return (
            geometry is Geometry.POLAR
            or geometry is Geometry.CYLINDRICAL
            or geometry is Geometry.SPHERICAL
        )

    def setup_fluid_fields(self):
        pass

    def setup_fluid_index_fields(self):
        # Now we get all our index types and set up aliases to them
        if self.ds is None:
            return
        index_fields = {f for _, f in self if _ == "index"}
        for ftype in self.ds.fluid_types:
            if ftype in ("index", "deposit"):
                continue
            for f in index_fields:
                if (ftype, f) in self:
                    continue
                self.alias((ftype, f), ("index", f))

    def setup_particle_fields(self, ptype, ftype="gas", num_neighbors=64):
        skip_output_units = ("code_length",)
        for f, (units, aliases, dn) in sorted(self.known_particle_fields):
            units = self.ds.field_units.get((ptype, f), units)
            output_units = units
            if (
                f in aliases or ptype not in self.ds.particle_types_raw
            ) and units not in skip_output_units:
                u = Unit(units, registry=self.ds.unit_registry)
                if u.dimensions is not dimensionless:
                    output_units = str(self.ds.unit_system[u.dimensions])
            if (ptype, f) not in self.field_list:
                continue
            self.add_output_field(
                (ptype, f),
                sampling_type="particle",
                units=units,
                display_name=dn,
                output_units=output_units,
            )
            for alias in aliases:
                self.alias((ptype, alias), (ptype, f), units=output_units)

        # We'll either have particle_position or particle_position_[xyz]
        if (ptype, "particle_position") in self.field_list or (
            ptype,
            "particle_position",
        ) in self.field_aliases:
            particle_scalar_functions(
                ptype, "particle_position", "particle_velocity", self
            )
        else:
            # We need to check to make sure that there's a "known field" that
            # overlaps with one of the vector fields.  For instance, if we are
            # in the Stream frontend, and we have a set of scalar position
            # fields, they will overlap with -- and be overridden by -- the
            # "known" vector field that the frontend creates.  So the easiest
            # thing to do is to simply remove the on-disk field (which doesn't
            # exist) and replace it with a derived field.
            if (ptype, "particle_position") in self and self[
                ptype, "particle_position"
            ]._function == NullFunc:
                self.pop((ptype, "particle_position"))
            particle_vector_functions(
                ptype,
                [f"particle_position_{ax}" for ax in "xyz"],
                [f"particle_velocity_{ax}" for ax in "xyz"],
                self,
            )
        particle_deposition_functions(ptype, "particle_position", "particle_mass", self)
        standard_particle_fields(self, ptype)
        # Now we check for any leftover particle fields
        for field in sorted(self.field_list):
            if field in self:
                continue
            if not isinstance(field, tuple):
                raise RuntimeError
            if field[0] not in self.ds.particle_types:
                continue
            units = self.ds.field_units.get(field, None)
            if units is None:
                try:
                    units = ytcfg.get("fields", *field, "units")
                except KeyError:
                    units = ""
            self.add_output_field(
                field,
                sampling_type="particle",
                units=units,
            )
        self.setup_smoothed_fields(ptype, num_neighbors=num_neighbors, ftype=ftype)

    def setup_extra_union_fields(self, ptype="all"):
        if ptype != "all":
            raise RuntimeError(
                "setup_extra_union_fields is currently"
                + 'only enabled for particle type "all".'
            )
        for units, field in self.extra_union_fields:
            add_union_field(self, ptype, field, units)

    def setup_smoothed_fields(self, ptype, num_neighbors=64, ftype="gas"):
        # We can in principle compute this, but it is not yet implemented.
        if (ptype, "density") not in self or not hasattr(self.ds, "_sph_ptypes"):
            return
        new_aliases = []
        for ptype2, alias_name in list(self):
            if ptype2 != ptype:
                continue
            if alias_name not in sph_whitelist_fields:
                if alias_name.startswith("particle_"):
                    pass
                else:
                    continue
            uni_alias_name = alias_name
            if "particle_position_" in alias_name:
                uni_alias_name = alias_name.replace("particle_position_", "")
            elif "particle_" in alias_name:
                uni_alias_name = alias_name.replace("particle_", "")
            new_aliases.append(
                (
                    (ftype, uni_alias_name),
                    (ptype, alias_name),
                )
            )
            if "particle_position_" in alias_name:
                new_aliases.append(
                    (
                        (ftype, alias_name),
                        (ptype, alias_name),
                    )
                )
            new_aliases.append(
                (
                    (ptype, uni_alias_name),
                    (ptype, alias_name),
                )
            )
            for alias, source in new_aliases:
                self.alias(alias, source)
        self.alias((ftype, "particle_position"), (ptype, "particle_position"))
        self.alias((ftype, "particle_mass"), (ptype, "particle_mass"))

    # Collect the names for all aliases if geometry is curvilinear
    def get_aliases_gallery(self) -> List[FieldName]:
        aliases_gallery: List[FieldName] = []
        known_other_fields = dict(self.known_other_fields)

        if self.ds is None:
            return aliases_gallery

        geometry: Geometry = self.ds.geometry
        if (
            geometry is Geometry.POLAR
            or geometry is Geometry.CYLINDRICAL
            or geometry is Geometry.SPHERICAL
        ):
            aliases: List[FieldName]
            for field in sorted(self.field_list):
                if field[0] in self.ds.particle_types:
                    continue
                args = known_other_fields.get(field[1], ("", [], None))
                units, aliases, display_name = args
                aliases_gallery.extend(aliases)
        elif (
            geometry is Geometry.CARTESIAN
            or geometry is Geometry.GEOGRAPHIC
            or geometry is Geometry.INTERNAL_GEOGRAPHIC
            or geometry is Geometry.SPECTRAL_CUBE
        ):
            # nothing to do
            pass
        else:
            assert_never(geometry)
        return aliases_gallery

    def setup_fluid_aliases(self, ftype: FieldType = "gas") -> None:
        known_other_fields = dict(self.known_other_fields)

        # For non-Cartesian geometry, convert alias of vector fields to
        # curvilinear coordinates
        aliases_gallery = self.get_aliases_gallery()

        for field in sorted(self.field_list):
            if not isinstance(field, tuple) or len(field) != 2:
                raise RuntimeError
            if field[0] in self.ds.particle_types:
                continue
            args = known_other_fields.get(field[1], None)
            if args is not None:
                units, aliases, display_name = args
            else:
                try:
                    node = ytcfg.get("fields", *field).as_dict()
                except KeyError:
                    node = {}

                units = node.get("units", "")
                aliases = node.get("aliases", [])
                display_name = node.get("display_name", None)

            # We allow field_units to override this.  First we check if the
            # field *name* is in there, then the field *tuple*.
            units = self.ds.field_units.get(field[1], units)
            units = self.ds.field_units.get(field, units)
            self.add_output_field(
                field, sampling_type="cell", units=units, display_name=display_name
            )
            axis_names = self.ds.coordinates.axis_order
            geometry: Geometry = self.ds.geometry
            for alias in aliases:
                if (
                    geometry is Geometry.POLAR
                    or geometry is Geometry.CYLINDRICAL
                    or geometry is Geometry.SPHERICAL
                ):
                    if alias[-2:] not in ["_x", "_y", "_z"]:
                        to_convert = False
                    else:
                        for suffix in ["x", "y", "z"]:
                            if f"{alias[:-2]}_{suffix}" not in aliases_gallery:
                                to_convert = False
                                break
                        to_convert = True
                    if to_convert:
                        if alias[-2:] == "_x":
                            alias = f"{alias[:-2]}_{axis_names[0]}"
                        elif alias[-2:] == "_y":
                            alias = f"{alias[:-2]}_{axis_names[1]}"
                        elif alias[-2:] == "_z":
                            alias = f"{alias[:-2]}_{axis_names[2]}"
                elif (
                    geometry is Geometry.CARTESIAN
                    or geometry is Geometry.GEOGRAPHIC
                    or geometry is Geometry.INTERNAL_GEOGRAPHIC
                    or geometry is Geometry.SPECTRAL_CUBE
                ):
                    # nothing to do
                    pass
                else:
                    assert_never(geometry)
                self.alias((ftype, alias), field)

    @staticmethod
    def _sanitize_sampling_type(sampling_type: str) -> str:
        """Detect conflicts between deprecated and new parameters to specify the
        sampling type in a new field.

        This is a helper function to add_field methods.

        Parameters
        ----------
        sampling_type : str
            One of "cell", "particle" or "local" (case insensitive)

        Raises
        ------
        ValueError
            For unsupported values in sampling_type
        """
        if not isinstance(sampling_type, str):
            raise TypeError("sampling_type should be a string.")

        sampling_type = sampling_type.lower()
        acceptable_samplings = ("cell", "particle", "local")
        if sampling_type not in acceptable_samplings:
            raise ValueError(
                f"Received invalid sampling type {sampling_type!r}. "
                f"Expected any of {acceptable_samplings}"
            )
        return sampling_type

    def add_field(
        self,
        name: FieldKey,
        function: Callable,
        sampling_type: str,
        *,
        alias: Optional[DerivedField] = None,
        force_override: bool = False,
        **kwargs,
    ) -> None:
        """
        Add a new field, along with supplemental metadata, to the list of
        available fields.  This respects a number of arguments, all of which
        are passed on to the constructor for
        :class:`~yt.data_objects.api.DerivedField`.

        Parameters
        ----------

        name : tuple[str, str]
           field (or particle) type, field name
        function : callable
           A function handle that defines the field.  Should accept
           arguments (field, data)
        sampling_type: str
           "cell" or "particle" or "local"
        force_override: bool
           If False (default), an error will be raised if a field of the same name already exists.
        alias: DerivedField (optional):
           existing field to be aliased
        units : str
           A plain text string encoding the unit.  Powers must be in
           python syntax (** instead of ^). If set to "auto" the units
           will be inferred from the return value of the field function.
        take_log : bool
           Describes whether the field should be logged
        validators : list
           A list of :class:`FieldValidator` objects
        vector_field : bool
           Describes the dimensionality of the field.  Currently unused.
        display_name : str
           A name used in the plots

        """
        # Handle the case where the field has already been added.
        if not force_override and name in self:
            return

        kwargs.setdefault("ds", self.ds)

        sampling_type = self._sanitize_sampling_type(sampling_type)

        if (
            not isinstance(name, str)
            and obj_length(name) == 2
            and all(isinstance(e, str) for e in name)
        ):
            self[name] = DerivedField(
                name, sampling_type, function, alias=alias, **kwargs
            )
        else:
            raise ValueError(f"Expected name to be a tuple[str, str], got {name}")

    def load_all_plugins(self, ftype: Optional[str] = "gas") -> None:
        if ftype is None:
            return
        mylog.debug("Loading field plugins for field type: %s.", ftype)
        loaded = []
        for n in sorted(field_plugins):
            loaded += self.load_plugin(n, ftype)
            only_on_root(mylog.debug, "Loaded %s (%s new fields)", n, len(loaded))
        self.find_dependencies(loaded)

    def load_plugin(
        self,
        plugin_name: FunctionName,
        ftype: FieldType = "gas",
        skip_check: bool = False,
    ):
        f = field_plugins[plugin_name]
        orig = set(self.items())
        f(self, ftype, slice_info=self.slice_info)
        loaded = [n for n, v in set(self.items()).difference(orig)]
        return loaded

    def find_dependencies(self, loaded):
        deps, unavailable = self.check_derived_fields(loaded)
        self.ds.field_dependencies.update(deps)
        # Note we may have duplicated
        dfl = set(self.ds.derived_field_list).union(deps.keys())
        self.ds.derived_field_list = sorted(dfl)
        return loaded, unavailable

    def add_output_field(self, name, sampling_type, **kwargs):
        if name[1] == "density":
            if name in self:
                # this should not happen, but it does
                # it'd be best to raise an error here but
                # it may take a while to cleanup internal issues
                return
        kwargs.setdefault("ds", self.ds)
        self[name] = DerivedField(name, sampling_type, NullFunc, **kwargs)

    def alias(
        self,
        alias_name: FieldKey,
        original_name: FieldKey,
        units: Optional[str] = None,
        deprecate: Optional[Tuple[str, Optional[str]]] = None,
    ):
        """
        Alias one field to another field.

        Parameters
        ----------
        alias_name : tuple[str, str]
            The new field name.
        original_name : tuple[str, str]
            The field to be aliased.
        units : str
           A plain text string encoding the unit.  Powers must be in
           python syntax (** instead of ^). If set to "auto" the units
           will be inferred from the return value of the field function.
        deprecate : tuple[str, str | None] | None
            If this is set, then the tuple contains two string version
            numbers: the first marking the version when the field was
            deprecated, and the second marking when the field will be
            removed.
        """
        if original_name not in self:
            return
        if units is None:
            # We default to CGS here, but in principle, this can be pluggable
            # as well.

            # self[original_name].units may be set to `None` at this point
            # to signal that units should be autoset later
            oru = self[original_name].units
            if oru is None:
                units = None
            else:
                u = Unit(oru, registry=self.ds.unit_registry)
                if u.dimensions is not dimensionless:
                    units = str(self.ds.unit_system[u.dimensions])
                else:
                    units = oru

        self.field_aliases[alias_name] = original_name
        function = TranslationFunc(original_name)
        if deprecate is not None:
            self.add_deprecated_field(
                alias_name,
                function=function,
                sampling_type=self[original_name].sampling_type,
                display_name=self[original_name].display_name,
                units=units,
                since=deprecate[0],
                removal=deprecate[1],
                ret_name=original_name,
            )
        else:
            self.add_field(
                alias_name,
                function=function,
                sampling_type=self[original_name].sampling_type,
                display_name=self[original_name].display_name,
                units=units,
                alias=self[original_name],
            )

    def add_deprecated_field(
        self,
        name,
        function,
        sampling_type,
        since,
        removal=None,
        ret_name=None,
        **kwargs,
    ):
        """
        Add a new field which is deprecated, along with supplemental metadata,
        to the list of available fields.  This respects a number of arguments,
        all of which are passed on to the constructor for
        :class:`~yt.data_objects.api.DerivedField`.

        Parameters
        ----------
        name : str
           is the name of the field.
        function : callable
           A function handle that defines the field.  Should accept
           arguments (field, data)
        sampling_type : str
           "cell" or "particle" or "local"
        since : str
            The version string marking when this field was deprecated.
        removal : str
            The version string marking when this field will be removed.
        ret_name : str
            The name of the field which will actually be returned, used
            only by :meth:`~yt.fields.field_info_container.FieldInfoContainer.alias`.
        units : str
           A plain text string encoding the unit.  Powers must be in
           python syntax (** instead of ^). If set to "auto" the units
           will be inferred from the return value of the field function.
        take_log : bool
           Describes whether the field should be logged
        validators : list
           A list of :class:`FieldValidator` objects
        vector_field : bool
           Describes the dimensionality of the field.  Currently unused.
        display_name : str
           A name used in the plots
        """
        if ret_name is None:
            ret_name = name
        self.add_field(
            name,
            function=DeprecatedFieldFunc(ret_name, function, since, removal),
            sampling_type=sampling_type,
            **kwargs,
        )

    def has_key(self, key):
        # This gets used a lot
        if key in self:
            return True
        if self.fallback is None:
            return False
        return key in self.fallback

    def __missing__(self, key):
        if self.fallback is None:
            raise KeyError(f"No field named {key}")
        return self.fallback[key]

    @classmethod
    def create_with_fallback(cls, fallback, name=""):
        obj = cls()
        obj.fallback = fallback
        obj.name = name
        return obj

    def __contains__(self, key):
        if super().__contains__(key):
            return True
        if self.fallback is None:
            return False
        return key in self.fallback

    def __iter__(self):
        yield from super().__iter__()
        if self.fallback is not None:
            yield from self.fallback

    def keys(self):
        keys = super().keys()
        if self.fallback:
            keys += list(self.fallback.keys())
        return keys

    def check_derived_fields(self, fields_to_check=None):
        # The following exceptions lists were obtained by expanding an
        # all-catching `except Exception`.
        # We define
        # - a blacklist (exceptions that we know should be caught)
        # - a whitelist (exceptions that should be handled)
        # - a greylist (exceptions that may be covering bugs but should be checked)
        # See https://github.com/yt-project/yt/issues/2853
        # in the long run, the greylist should be removed
        blacklist = ()
        whitelist = (NotImplementedError,)
        greylist = (
            YTFieldNotFound,
            YTDomainOverflow,
            YTCoordinateNotImplemented,
            NeedsConfiguration,
            TypeError,
            ValueError,
            IndexError,
            AttributeError,
            KeyError,
            # code smells -> those are very likely bugs
            UnitConversionError,  # solved in GH PR 2897 ?
            # RecursionError is clearly a bug, and was already solved once
            # in GH PR 2851
            RecursionError,
        )

        deps = {}
        unavailable = []
        fields_to_check = fields_to_check or list(self.keys())
        for field in fields_to_check:
            fi = self[field]
            try:
                # fd: field detector
                fd = fi.get_dependencies(ds=self.ds)
            except blacklist as err:
                print(f"{err.__class__} raised for field {field}")
                raise SystemExit(1) from err
            except (*whitelist, *greylist) as e:
                if field in self._show_field_errors:
                    raise
                if not isinstance(e, YTFieldNotFound):
                    # if we're doing field tests, raise an error
                    # see yt.fields.tests.test_fields
                    if hasattr(self.ds, "_field_test_dataset"):
                        raise
                    mylog.debug(
                        "Raises %s during field %s detection.", str(type(e)), field
                    )
                self.pop(field)
                continue
            # This next bit checks that we can't somehow generate everything.
            # We also manually update the 'requested' attribute
            missing = not all(f in self.field_list for f in fd.requested)
            if missing:
                self.pop(field)
                unavailable.append(field)
                continue
            fd.requested = set(fd.requested)
            deps[field] = fd
            mylog.debug("Succeeded with %s (needs %s)", field, fd.requested)

        # now populate the derived field list with results
        # this violates isolation principles and should be refactored
        dfl = set(self.ds.derived_field_list).union(deps.keys())
        dfl = sorted(dfl)

        if not hasattr(self.ds.index, "meshes"):
            # the meshes attribute characterizes an unstructured-mesh data structure

            # ideally this filtering should not be required
            # and this could maybe be handled in fi.get_dependencies
            # but it's a lot easier to do here

            filtered_dfl = []
            for field in dfl:
                try:
                    ftype, fname = field
                    if "vertex" in fname:
                        continue
                except ValueError:
                    # in very rare cases, there can a field represented by a single
                    # string, like "emissivity"
                    # this try block _should_ be removed and the error fixed upstream
                    # for reference, a test that would break is
                    # yt/data_objects/tests/test_fluxes.py::ExporterTests
                    pass
                filtered_dfl.append(field)
            dfl = filtered_dfl

        self.ds.derived_field_list = dfl
        self._set_linear_fields()
        return deps, unavailable

    def _set_linear_fields(self):
        """
        Sets which fields use linear as their default scaling in Profiles and
        PhasePlots. Default for all fields is set to log, so this sets which
        are linear.  For now, set linear to geometric fields: position and
        velocity coordinates.
        """
        non_log_prefixes = ("", "velocity_", "particle_position_", "particle_velocity_")
        coords = ("x", "y", "z")
        non_log_fields = [
            prefix + coord for prefix in non_log_prefixes for coord in coords
        ]
        for field in self.ds.derived_field_list:
            if field[1] in non_log_fields:
                self[field].take_log = False



import numpy as np

from yt.data_objects.api import ImageArray
from yt.funcs import is_sequence, mylog
from yt.units.unit_object import Unit  # type: ignore
from yt.utilities.lib.partitioned_grid import PartitionedGrid
from yt.utilities.lib.pixelization_routines import (
    normalization_2d_utility,
    off_axis_projection_SPH,
)

from .render_source import KDTreeVolumeSource
from .scene import Scene
from .transfer_functions import ProjectionTransferFunction
from .utils import data_source_or_all


def off_axis_projection(
    data_source,
    center,
    normal_vector,
    width,
    resolution,
    item,
    weight=None,
    volume=None,
    no_ghost=False,
    interpolated=False,
    north_vector=None,
    num_threads=1,
    method="integrate",
):
    r"""Project through a dataset, off-axis, and return the image plane.

    This function will accept the necessary items to integrate through a volume
    at an arbitrary angle and return the integrated field of view to the user.
    Note that if a weight is supplied, it will multiply the pre-interpolated
    values together, then create cell-centered values, then interpolate within
    the cell to conduct the integration.

    Parameters
    ----------
    data_source : ~yt.data_objects.static_output.Dataset
                  or ~yt.data_objects.data_containers.YTSelectionDataContainer
        This is the dataset or data object to volume render.
    center : array_like
        The current 'center' of the view port -- the focal point for the
        camera.
    normal_vector : array_like
        The vector between the camera position and the center.
    width : float or list of floats
        The current width of the image.  If a single float, the volume is
        cubical, but if not, it is left/right, top/bottom, front/back
    resolution : int or list of ints
        The number of pixels in each direction.
    item: string
        The field to project through the volume
    weight : optional, default None
        If supplied, the field will be pre-multiplied by this, then divided by
        the integrated value of this field.  This returns an average rather
        than a sum.
    volume : `yt.extensions.volume_rendering.AMRKDTree`, optional
        The volume to ray cast through.  Can be specified for finer-grained
        control, but otherwise will be automatically generated.
    no_ghost: bool, optional
        Optimization option.  If True, homogenized bricks will
        extrapolate out from grid instead of interpolating from
        ghost zones that have to first be calculated.  This can
        lead to large speed improvements, but at a loss of
        accuracy/smoothness in resulting image.  The effects are
        less notable when the transfer function is smooth and
        broad. Default: True
    interpolated : optional, default False
        If True, the data is first interpolated to vertex-centered data,
        then tri-linearly interpolated along the ray. Not suggested for
        quantitative studies.
    north_vector : optional, array_like, default None
        A vector that, if specified, restricts the orientation such that the
        north vector dotted into the image plane points "up". Useful for rotations
    num_threads: integer, optional, default 1
        Use this many OpenMP threads during projection.
    method : string
        The method of projection.  Valid methods are:

        "integrate" with no weight_field specified : integrate the requested
        field along the line of sight.

        "integrate" with a weight_field specified : weight the requested
        field by the weighting field and integrate along the line of sight.

        "sum" : This method is the same as integrate, except that it does not
        multiply by a path length when performing the integration, and is
        just a straight summation of the field along the given axis. WARNING:
        This should only be used for uniform resolution grid datasets, as other
        datasets may result in unphysical images.
        or camera movements.
    Returns
    -------
    image : array
        An (N,N) array of the final integrated values, in float64 form.

    Examples
    --------

    >>> image = off_axis_projection(
    ...     ds,
    ...     [0.5, 0.5, 0.5],
    ...     [0.2, 0.3, 0.4],
    ...     0.2,
    ...     N,
    ...     ("gas", "temperature"),
    ...     ("gas", "density"),
    ... )
    >>> write_image(np.log10(image), "offaxis.png")

    """
    if method not in ("integrate", "sum"):
        raise NotImplementedError(
            "Only 'integrate' or 'sum' methods are valid for off-axis-projections"
        )

    if interpolated:
        raise NotImplementedError(
            "Only interpolated=False methods are currently implemented "
            "for off-axis-projections"
        )

    data_source = data_source_or_all(data_source)

    item = data_source._determine_fields([item])[0]

    # Assure vectors are numpy arrays as expected by cython code
    normal_vector = np.array(normal_vector, dtype="float64")
    if north_vector is not None:
        north_vector = np.array(north_vector, dtype="float64")
    # Add the normal as a field parameter to the data source
    # so line of sight fields can use it
    data_source.set_field_parameter("axis", normal_vector)

    # Sanitize units
    if not hasattr(center, "units"):
        center = data_source.ds.arr(center, "code_length")
    if not hasattr(width, "units"):
        width = data_source.ds.arr(width, "code_length")

    if hasattr(data_source.ds, "_sph_ptypes"):
        if method != "integrate":
            raise NotImplementedError("SPH Only allows 'integrate' method")

        sph_ptypes = data_source.ds._sph_ptypes
        fi = data_source.ds.field_info[item]

        raise_error = False

        ptype = sph_ptypes[0]
        ppos = [f"particle_position_{ax}" for ax in "xyz"]
        # Assure that the field we're trying to off-axis project
        # has a field type as the SPH particle type or if the field is an
        # alias to an SPH field or is a 'gas' field
        if item[0] in data_source.ds.known_filters:
            if item[0] not in sph_ptypes:
                raise_error = True
            else:
                ptype = item[0]
                ppos = ["x", "y", "z"]
        elif fi.is_alias:
            if fi.alias_name[0] not in sph_ptypes:
                raise_error = True
            elif item[0] != "gas":
                ptype = item[0]
        else:
            if fi.name[0] not in sph_ptypes and fi.name[0] != "gas":
                raise_error = True

        if raise_error:
            raise RuntimeError(
                "Can only perform off-axis projections for SPH fields, "
                f"Received {item!r}"
            )

        normal = np.array(normal_vector)
        normal = normal / np.linalg.norm(normal)

        # If north_vector is None, we set the default here.
        # This is chosen so that if normal_vector is one of the
        # cartesian coordinate axes, the projection will match
        # the corresponding on-axis projection.
        if north_vector is None:
            vecs = np.identity(3)
            t = np.cross(vecs, normal).sum(axis=1)
            ax = t.argmax()
            east_vector = np.cross(vecs[ax, :], normal).ravel()
            north = np.cross(normal, east_vector).ravel()
        else:
            north = np.array(north_vector)
            north = north / np.linalg.norm(north)
            east_vector = np.cross(north, normal).ravel()

        # if weight is None:
        buf = np.zeros((resolution[0], resolution[1]), dtype="float64")

        x_min = center[0] - width[0] / 2
        x_max = center[0] + width[0] / 2
        y_min = center[1] - width[1] / 2
        y_max = center[1] + width[1] / 2
        z_min = center[2] - width[2] / 2
        z_max = center[2] + width[2] / 2
        finfo = data_source.ds.field_info[item]
        ounits = finfo.output_units
        bounds = [x_min, x_max, y_min, y_max, z_min, z_max]

        if weight is None:
            for chunk in data_source.chunks([], "io"):
                off_axis_projection_SPH(
                    chunk[ptype, ppos[0]].to("code_length").d,
                    chunk[ptype, ppos[1]].to("code_length").d,
                    chunk[ptype, ppos[2]].to("code_length").d,
                    chunk[ptype, "mass"].to("code_mass").d,
                    chunk[ptype, "density"].to("code_density").d,
                    chunk[ptype, "smoothing_length"].to("code_length").d,
                    bounds,
                    center.to("code_length").d,
                    width.to("code_length").d,
                    chunk[item].in_units(ounits),
                    buf,
                    normal_vector,
                    north,
                )

            # Assure that the path length unit is in the default length units
            # for the dataset by scaling the units of the smoothing length,
            # which in the above calculation is set to be code_length
            path_length_unit = Unit(
                "code_length", registry=data_source.ds.unit_registry
            )
            default_path_length_unit = data_source.ds.unit_system["length"]
            buf *= data_source.ds.quan(1, path_length_unit).in_units(
                default_path_length_unit
            )
            item_unit = data_source.ds._get_field_info(item).units
            item_unit = Unit(item_unit, registry=data_source.ds.unit_registry)
            funits = item_unit * default_path_length_unit

        else:
            # if there is a weight field, take two projections:
            # one of field*weight, the other of just weight, and divide them
            weight_buff = np.zeros((resolution[0], resolution[1]), dtype="float64")
            wounits = data_source.ds.field_info[weight].output_units

            for chunk in data_source.chunks([], "io"):
                off_axis_projection_SPH(
                    chunk[ptype, ppos[0]].to("code_length").d,
                    chunk[ptype, ppos[1]].to("code_length").d,
                    chunk[ptype, ppos[2]].to("code_length").d,
                    chunk[ptype, "mass"].to("code_mass").d,
                    chunk[ptype, "density"].to("code_density").d,
                    chunk[ptype, "smoothing_length"].to("code_length").d,
                    bounds,
                    center.to("code_length").d,
                    width.to("code_length").d,
                    chunk[item].in_units(ounits),
                    buf,
                    normal_vector,
                    north,
                    weight_field=chunk[weight].in_units(wounits),
                )

            for chunk in data_source.chunks([], "io"):
                off_axis_projection_SPH(
                    chunk[ptype, ppos[0]].to("code_length").d,
                    chunk[ptype, ppos[1]].to("code_length").d,
                    chunk[ptype, ppos[2]].to("code_length").d,
                    chunk[ptype, "mass"].to("code_mass").d,
                    chunk[ptype, "density"].to("code_density").d,
                    chunk[ptype, "smoothing_length"].to("code_length").d,
                    bounds,
                    center.to("code_length").d,
                    width.to("code_length").d,
                    chunk[weight].to(wounits),
                    weight_buff,
                    normal_vector,
                    north,
                )

            normalization_2d_utility(buf, weight_buff)
            item_unit = data_source.ds._get_field_info(item).units
            item_unit = Unit(item_unit, registry=data_source.ds.unit_registry)
            funits = item_unit

        myinfo = {
            "field": item,
            "east_vector": east_vector,
            "north_vector": north_vector,
            "normal_vector": normal_vector,
            "width": width,
            "units": funits,
            "type": "SPH smoothed projection",
        }

        return ImageArray(
            buf, funits, registry=data_source.ds.unit_registry, info=myinfo
        )

    sc = Scene()
    data_source.ds.index
    if item is None:
        field = data_source.ds.field_list[0]
        mylog.info("Setting default field to %s", field.__repr__())

    funits = data_source.ds._get_field_info(item).units

    vol = KDTreeVolumeSource(data_source, item)
    vol.num_threads = num_threads
    if weight is None:
        vol.set_field(item)
    else:
        # This is a temporary field, which we will remove at the end.
        weightfield = ("index", "temp_weightfield")

        def _make_wf(f, w):
            def temp_weightfield(field, data):
                tr = data[f].astype("float64") * data[w]
                return tr.d

            return temp_weightfield

        data_source.ds.field_info.add_field(
            weightfield,
            sampling_type="cell",
            function=_make_wf(item, weight),
            units="",
        )
        # Now we have to tell the dataset to add it and to calculate
        # its dependencies..
        deps, _ = data_source.ds.field_info.check_derived_fields([weightfield])
        data_source.ds.field_dependencies.update(deps)
        vol.set_field(weightfield)
        vol.set_weight_field(weight)
    ptf = ProjectionTransferFunction()
    vol.set_transfer_function(ptf)
    camera = sc.add_camera(data_source)
    camera.set_width(width)
    if not is_sequence(resolution):
        resolution = [resolution] * 2
    camera.resolution = resolution
    if not is_sequence(width):
        width = data_source.ds.arr([width] * 3)
    normal = np.array(normal_vector)
    normal = normal / np.linalg.norm(normal)

    camera.position = center - width[2] * normal
    camera.focus = center

    # If north_vector is None, we set the default here.
    # This is chosen so that if normal_vector is one of the
    # cartesian coordinate axes, the projection will match
    # the corresponding on-axis projection.
    if north_vector is None:
        vecs = np.identity(3)
        t = np.cross(vecs, normal).sum(axis=1)
        ax = t.argmax()
        east_vector = np.cross(vecs[ax, :], normal).ravel()
        north = np.cross(normal, east_vector).ravel()
    else:
        north = np.array(north_vector)
        north = north / np.linalg.norm(north)
    camera.switch_orientation(normal, north)

    sc.add_source(vol)

    vol.set_sampler(camera, interpolated=False)
    assert vol.sampler is not None

    fields = [vol.field]
    if vol.weight_field is not None:
        fields.append(vol.weight_field)

    mylog.debug("Casting rays")

    for grid, mask in data_source.blocks:
        data = []
        for f in fields:
            # strip units before multiplying by mask for speed
            grid_data = grid[f]
            units = grid_data.units
            data.append(data_source.ds.arr(grid_data.d * mask, units, dtype="float64"))
        pg = PartitionedGrid(
            grid.id,
            data,
            mask.astype("uint8"),
            grid.LeftEdge,
            grid.RightEdge,
            grid.ActiveDimensions.astype("int64"),
        )
        grid.clear_data()
        vol.sampler(pg, num_threads=num_threads)

    image = vol.finalize_image(camera, vol.sampler.aimage)
    image = ImageArray(
        image, funits, registry=data_source.ds.unit_registry, info=image.info
    )

    if weight is not None:
        data_source.ds.field_info.pop(("index", "temp_weightfield"))

    if method == "integrate":
        if weight is None:
            dl = width[2].in_units(data_source.ds.unit_system["length"])
            image *= dl
        else:
            mask = image[:, :, 1] == 0
            image[:, :, 0] /= image[:, :, 1]
            image[mask] = 0

    return image[:, :, 0]



from io import BytesIO

from PIL import Image


def call_png_write_png(buffer, fileobj, dpi):
    Image.fromarray(buffer).save(fileobj, dpi=(dpi, dpi), format="png")


def write_png(buffer, filename, dpi=100):
    with open(filename, "wb") as fileobj:
        call_png_write_png(buffer, fileobj, dpi)


def write_png_to_string(buffer, dpi=100):
    fileobj = BytesIO()
    call_png_write_png(buffer, fileobj, dpi)
    png_str = fileobj.getvalue()
    fileobj.close()
    return png_str



import abc
import os
from typing import List, Optional, Set, Tuple, Type

from yt._typing import FieldKey
from yt.config import ytcfg
from yt.funcs import mylog
from yt.utilities.cython_fortran_utils import FortranFile

from .field_handlers import HandlerMixin
from .io import _read_part_file_descriptor

PARTICLE_HANDLERS: Set[Type["ParticleFileHandler"]] = set()


def get_particle_handlers():
    return PARTICLE_HANDLERS


def register_particle_handler(ph):
    PARTICLE_HANDLERS.add(ph)


class ParticleFileHandler(abc.ABC, HandlerMixin):
    """
    Abstract class to handle particles in RAMSES. Each instance
    represents a single file (one domain).

    To add support to a new particle file, inherit from this class and
    implement all functions containing a `NotImplementedError`.

    See `SinkParticleFileHandler` for an example implementation."""

    _file_type = "particle"

    # These properties are static properties
    ptype: Optional[str] = None  # The name to give to the particle type
    fname: Optional[str] = None  # The name of the file(s).
    file_descriptor: Optional[str] = None  # The name of the file descriptor (if any)

    attrs: Tuple[Tuple[str, int, str], ...]  # The attributes of the header
    known_fields: Optional[
        List[FieldKey]
    ] = None  # A list of tuple containing the field name and its type
    config_field: Optional[str] = None  # Name of the config section (if any)

    # These properties are computed dynamically
    field_offsets = None  # Mapping from field to offset in file
    field_types = (
        None  # Mapping from field to the type of the data (float, integer, ...)
    )
    local_particle_count = None  # The number of particle in the domain

    def __init_subclass__(cls, *args, **kwargs):
        """
        Registers subclasses at creation.
        """
        super().__init_subclass__(*args, **kwargs)

        if cls.ptype is not None:
            register_particle_handler(cls)

        cls._unique_registry = {}
        return cls

    def __init__(self, domain):
        self.setup_handler(domain)

        # Attempt to read the list of fields from the config file
        if self.config_field and ytcfg.has_section(self.config_field):
            cfg = ytcfg.get(self.config_field, "fields")
            known_fields = []
            for c in (_.strip() for _ in cfg.split("\n") if _.strip() != ""):
                field, field_type = (_.strip() for _ in c.split(","))
                known_fields.append((field, field_type))
            self.known_fields = known_fields

    @abc.abstractmethod
    def read_header(self):
        """
        This function is called once per file. It should:

        * read the header of the file and store any relevant information
        * detect the fields in the file
        * compute the offsets (location in the file) of each field

        It is in charge of setting `self.field_offsets` and `self.field_types`.

        * `field_offsets`: dictionary: tuple -> integer
            A dictionary that maps `(type, field_name)` to their
            location in the file (integer)
        * `field_types`: dictionary: tuple -> character
            A dictionary that maps `(type, field_name)` to their type
            (character), following Python's struct convention.
        """
        pass


class DefaultParticleFileHandler(ParticleFileHandler):
    ptype = "io"
    fname = "part_{iout:05d}.out{icpu:05d}"
    file_descriptor = "part_file_descriptor.txt"
    config_field = "ramses-particles"

    attrs = (
        ("ncpu", 1, "i"),
        ("ndim", 1, "i"),
        ("npart", 1, "i"),
        ("localseed", -1, "i"),
        ("nstar_tot", 1, "i"),
        ("mstar_tot", 1, "d"),
        ("mstar_lost", 1, "d"),
        ("nsink", 1, "i"),
    )

    known_fields = [
        ("particle_position_x", "d"),
        ("particle_position_y", "d"),
        ("particle_position_z", "d"),
        ("particle_velocity_x", "d"),
        ("particle_velocity_y", "d"),
        ("particle_velocity_z", "d"),
        ("particle_mass", "d"),
        ("particle_identity", "i"),
        ("particle_refinement_level", "i"),
    ]

    def read_header(self):
        if not self.exists:
            self.field_offsets = {}
            self.field_types = {}
            self.local_particle_count = 0
            return

        fd = FortranFile(self.fname)
        fd.seek(0, os.SEEK_END)
        flen = fd.tell()
        fd.seek(0)
        hvals = {}
        attrs = self.attrs
        hvals.update(fd.read_attrs(attrs))
        self.header = hvals
        self.local_particle_count = hvals["npart"]
        extra_particle_fields = self.ds._extra_particle_fields

        if self.has_descriptor:
            particle_fields = _read_part_file_descriptor(self.file_descriptor)
        else:
            particle_fields = list(self.known_fields)

            if extra_particle_fields is not None:
                particle_fields += extra_particle_fields

        if hvals["nstar_tot"] > 0 and extra_particle_fields is not None:
            particle_fields += [
                ("particle_birth_time", "d"),
                ("particle_metallicity", "d"),
            ]

        field_offsets = {}
        _pfields = {}

        ptype = self.ptype

        # Read offsets
        for field, vtype in particle_fields:
            if fd.tell() >= flen:
                break
            field_offsets[ptype, field] = fd.tell()
            _pfields[ptype, field] = vtype
            fd.skip(1)

        iextra = 0
        while fd.tell() < flen:
            iextra += 1
            field, vtype = ("particle_extra_field_%i" % iextra, "d")
            particle_fields.append((field, vtype))

            field_offsets[ptype, field] = fd.tell()
            _pfields[ptype, field] = vtype
            fd.skip(1)

        fd.close()

        if iextra > 0 and not self.ds._warned_extra_fields["io"]:
            mylog.warning(
                "Detected %s extra particle fields assuming kind "
                "`double`. Consider using the `extra_particle_fields` "
                "keyword argument if you have unexpected behavior.",
                iextra,
            )
            self.ds._warned_extra_fields["io"] = True

        self.field_offsets = field_offsets
        self.field_types = _pfields


class SinkParticleFileHandler(ParticleFileHandler):
    """Handle sink files"""

    ptype = "sink"
    fname = "sink_{iout:05d}.out{icpu:05d}"
    file_descriptor = "sink_file_descriptor.txt"
    config_field = "ramses-sink-particles"

    attrs = (("nsink", 1, "i"), ("nindsink", 1, "i"))

    known_fields = [
        ("particle_identifier", "i"),
        ("particle_mass", "d"),
        ("particle_position_x", "d"),
        ("particle_position_y", "d"),
        ("particle_position_z", "d"),
        ("particle_velocity_x", "d"),
        ("particle_velocity_y", "d"),
        ("particle_velocity_z", "d"),
        ("particle_birth_time", "d"),
        ("BH_real_accretion", "d"),
        ("BH_bondi_accretion", "d"),
        ("BH_eddington_accretion", "d"),
        ("BH_esave", "d"),
        ("gas_spin_x", "d"),
        ("gas_spin_y", "d"),
        ("gas_spin_z", "d"),
        ("BH_spin_x", "d"),
        ("BH_spin_y", "d"),
        ("BH_spin_z", "d"),
        ("BH_spin", "d"),
        ("BH_efficiency", "d"),
    ]

    def read_header(self):
        if not self.exists:
            self.field_offsets = {}
            self.field_types = {}
            self.local_particle_count = 0
            return
        fd = FortranFile(self.fname)
        fd.seek(0, os.SEEK_END)
        flen = fd.tell()
        fd.seek(0)
        hvals = {}
        # Read the header of the file
        attrs = self.attrs

        hvals.update(fd.read_attrs(attrs))
        self._header = hvals

        # This is somehow a trick here: we only want one domain to
        # be read, as ramses writes all the sinks in all the
        # domains. Here, we set the local_particle_count to 0 except
        # for the first domain to be red.
        if getattr(self.ds, "_sink_file_flag", False):
            self.local_particle_count = 0
        else:
            self.ds._sink_file_flag = True
            self.local_particle_count = hvals["nsink"]

        # Read the fields + add the sink properties
        if self.has_descriptor:
            fields = _read_part_file_descriptor(self.file_descriptor)
        else:
            fields = list(self.known_fields)

        # Note: this follows RAMSES convention.
        for i in range(self.ds.dimensionality * 2 + 1):
            for ilvl in range(self.ds.max_level + 1):
                fields.append((f"particle_prop_{ilvl}_{i}", "d"))

        field_offsets = {}
        _pfields = {}

        # Fill the fields, offsets and types
        self.fields = []
        for field, vtype in fields:
            self.fields.append(field)
            if fd.tell() >= flen:
                break
            field_offsets[self.ptype, field] = fd.tell()
            _pfields[self.ptype, field] = vtype
            fd.skip(1)
        self.field_offsets = field_offsets
        self.field_types = _pfields
        fd.close()



from collections import defaultdict

import numpy as np

from yt.frontends.open_pmd.misc import get_component, is_const_component
from yt.geometry.selection_routines import GridSelector
from yt.utilities.io_handler import BaseIOHandler


class IOHandlerOpenPMDHDF5(BaseIOHandler):
    _field_dtype = "float32"
    _dataset_type = "openPMD"

    def __init__(self, ds, *args, **kwargs):
        self.ds = ds
        self._handle = ds._handle
        self.base_path = ds.base_path
        self.meshes_path = ds.meshes_path
        self.particles_path = ds.particles_path
        self._array_fields = {}
        self._cached_ptype = ""

    def _fill_cache(self, ptype, index=0, offset=None):
        """Fills the particle position cache for the ``ptype``.

        Parameters
        ----------
        ptype : str
            The on-disk name of the particle species
        index : int, optional
        offset : int, optional
        """
        if str((ptype, index, offset)) not in self._cached_ptype:
            self._cached_ptype = str((ptype, index, offset))
            pds = self._handle[self.base_path + self.particles_path + "/" + ptype]
            axes = list(pds["position"].keys())
            if offset is None:
                if is_const_component(pds["position/" + axes[0]]):
                    offset = pds["position/" + axes[0]].attrs["shape"]
                else:
                    offset = pds["position/" + axes[0]].len()
            self.cache = np.empty((3, offset), dtype=np.float64)
            for i in np.arange(3):
                ax = "xyz"[i]
                if ax in axes:
                    np.add(
                        get_component(pds, "position/" + ax, index, offset),
                        get_component(pds, "positionOffset/" + ax, index, offset),
                        self.cache[i],
                    )
                else:
                    # Pad accordingly with zeros to make 1D/2D datasets compatible
                    # These have to be the same shape as the existing axes since that
                    # equals the number of particles
                    self.cache[i] = np.zeros(offset)

    def _read_particle_selection(self, chunks, selector, fields):
        """Read particle fields for particle species masked by a selection.

        Parameters
        ----------
        chunks
            A list of chunks
            A chunk is a list of grids
        selector
            A region (inside your domain) specifying which parts of the field
            you want to read. See [1] and [2]
        fields : array_like
            Tuples (ptype, pfield) representing a field

        Returns
        -------
        dict
            keys are tuples (ptype, pfield) representing a field
            values are (N,) ndarrays with data from that field
        """
        f = self._handle
        bp = self.base_path
        pp = self.particles_path
        ds = f[bp + pp]
        unions = self.ds.particle_unions
        chunks = list(chunks)  # chunks is a generator

        rv = {}
        ind = {}
        particle_count = {}
        ptf = defaultdict(list)  # ParticleTypes&Fields
        rfm = defaultdict(list)  # RequestFieldMapping

        for ptype, pname in fields:
            pfield = (ptype, pname)
            # Overestimate the size of all pfields so they include all particles
            # and shrink it later
            particle_count[pfield] = 0
            if ptype in unions:
                for pt in unions[ptype]:
                    particle_count[pfield] += self.ds.particle_type_counts[pt]
                    ptf[pt].append(pname)
                    rfm[pt, pname].append(pfield)
            else:
                particle_count[pfield] = self.ds.particle_type_counts[ptype]
                ptf[ptype].append(pname)
                rfm[pfield].append(pfield)
            rv[pfield] = np.empty((particle_count[pfield],), dtype=np.float64)
            ind[pfield] = 0

        for ptype in ptf:
            for chunk in chunks:
                for grid in chunk.objs:
                    if str(ptype) == "io":
                        species = list(ds.keys())[0]
                    else:
                        species = ptype
                    if species not in grid.ptypes:
                        continue
                    # read particle coords into cache
                    self._fill_cache(species, grid.pindex, grid.poffset)
                    mask = selector.select_points(
                        self.cache[0], self.cache[1], self.cache[2], 0.0
                    )
                    if mask is None:
                        continue
                    pds = ds[species]
                    for field in ptf[ptype]:
                        component = "/".join(field.split("_")[1:])
                        component = component.replace("positionCoarse", "position")
                        component = component.replace("-", "_")
                        data = get_component(pds, component, grid.pindex, grid.poffset)[
                            mask
                        ]
                        for request_field in rfm[(ptype, field)]:
                            rv[request_field][
                                ind[request_field] : ind[request_field] + data.shape[0]
                            ] = data
                            ind[request_field] += data.shape[0]

        for field in fields:
            rv[field] = rv[field][: ind[field]]

        return rv

    def _read_fluid_selection(self, chunks, selector, fields, size):
        """Reads given fields masked by a given selection.

        Parameters
        ----------
        chunks
            A list of chunks
            A chunk is a list of grids
        selector
            A region (inside your domain) specifying which parts of the field
            you want to read. See [1] and [2]
        fields : array_like
            Tuples (fname, ftype) representing a field
        size : int
            Size of the data to read

        Returns
        -------
        dict
            keys are tuples (ftype, fname) representing a field
            values are flat (``size``,) ndarrays with data from that field
        """
        f = self._handle
        bp = self.base_path
        mp = self.meshes_path
        ds = f[bp + mp]
        chunks = list(chunks)

        rv = {}
        ind = {}

        if isinstance(selector, GridSelector):
            if not (len(chunks) == len(chunks[0].objs) == 1):
                raise RuntimeError

        if size is None:
            size = sum(g.count(selector) for chunk in chunks for g in chunk.objs)
        for field in fields:
            rv[field] = np.empty(size, dtype=np.float64)
            ind[field] = 0

        for ftype, fname in fields:
            field = (ftype, fname)
            for chunk in chunks:
                for grid in chunk.objs:
                    mask = grid._get_selector_mask(selector)
                    if mask is None:
                        continue
                    component = fname.replace("_", "/").replace("-", "_")
                    if component.split("/")[0] not in grid.ftypes:
                        data = np.full(grid.ActiveDimensions, 0, dtype=np.float64)
                    else:
                        data = get_component(ds, component, grid.findex, grid.foffset)
                    # The following is a modified AMRGridPatch.select(...)
                    data.shape = (
                        mask.shape
                    )  # Workaround - casts a 2D (x,y) array to 3D (x,y,1)
                    count = grid.count(selector)
                    rv[field][ind[field] : ind[field] + count] = data[mask]
                    ind[field] += count

        for field in fields:
            rv[field] = rv[field][: ind[field]]
            rv[field].flatten()

        return rv



"""
This is a collection of helper functions to yt.load_sample
"""
import json
import re
import sys
from functools import lru_cache
from itertools import chain
from pathlib import Path
from typing import Optional, Union
from warnings import warn

from yt.config import ytcfg
from yt.funcs import mylog
from yt.utilities.on_demand_imports import (
    _pandas as pd,
    _pooch as pooch,
    _requests as requests,
)

if sys.version_info >= (3, 9):
    import importlib.resources as importlib_resources
else:
    import importlib_resources

num_exp = re.compile(r"\d*(\.\d*)?")
byte_unit_exp = re.compile(r"[KMGT]?B")


def _parse_byte_size(s: str):
    """
    Convert a string size specification to integer byte size.

    This function should be insensitive to case and whitespace.

    It doesn't always return an int, as a temporary measure to deal with missing
    or corrupted data in the registry. This should be fixed in the future.

    Examples
    --------
    # most of the following examples are adapted from
    # https://stackoverflow.com/a/31631711/5489622
    >>> _parse_byte_size(None)
    <NA>
    >>> from numpy import nan
    >>> _parse_byte_size(nan)
    <NA>
    >>> _parse_byte_size("1B")
    1
    >>> _parse_byte_size("1.00 KB")
    1024
    >>> _parse_byte_size("488.28 KB")
    500000
    >>> _parse_byte_size("1.00 MB")
    1049000
    >>> _parse_byte_size("47.68 MB")
    50000000
    >>> _parse_byte_size("1.00 GB")
    1074000000
    >>> _parse_byte_size("4.66 GB")
    5004000000
    >>> _parse_byte_size("1.00 TB")
    1100000000000
    >>> _parse_byte_size("4.55 TB")
    5003000000000
    """
    try:
        s = s.upper()
    except AttributeError:
        # input is not a string (likely a np.nan)
        return pd.NA

    match = re.search(num_exp, s)
    if match is None:
        raise ValueError
    val = float(match.group())

    match = re.search(byte_unit_exp, s)
    if match is None:
        raise ValueError
    unit = match.group()
    prefixes = ["B", "K", "M", "G", "T"]
    raw_res = val * 1024 ** prefixes.index(unit[0])
    return int(float(f"{raw_res:.3e}"))


def _get_sample_data_registry():
    return json.loads(
        importlib_resources.files("yt")
        .joinpath("sample_data_registry.json")
        .read_bytes()
    )


@lru_cache(maxsize=128)
def get_data_registry_table():
    """
    Load the sample data registry as a pandas.Dataframe instance.

    This function is considered experimental and is exposed for exploratory purposed.
    The output format is subject to change.

    The output of this function is cached so it will only generate one request per session.
    """
    # it would be nicer to have an actual api on the yt website server,
    # but this will do for now
    api_url = "https://raw.githubusercontent.com/yt-project/website/master/data/datafiles.json"

    response = requests.get(api_url)

    if not response.ok:
        raise RuntimeError(
            "Could not retrieve registry data. Please check your network setup."
        )

    website_json = response.json()
    # this dict follows this schema: {frontend_name: {flat dataframe-like}}

    columns = ["code", "filename", "size", "url", "description"]
    website_table = pd.concat(pd.DataFrame(d) for d in website_json.values())[columns]

    # add a int-type byte size column
    # note that we cast to pandas specific type "Int64" because we expect missing values
    # see https://pandas.pydata.org/pandas-docs/stable/user_guide/missing_data.html#integer-dtypes-and-missing-data
    website_table["byte_size"] = (
        website_table["size"].apply(_parse_byte_size).astype("Int64")
    )

    # normalize urls to match the local json
    website_table["url"] = website_table["url"].apply(
        lambda u: u.replace("http:", "https:")
    )

    # load local data
    pooch_table = pd.DataFrame(_get_sample_data_registry().values())

    # merge tables
    unified_table = website_table.merge(pooch_table, on="url", how="outer")

    # PR 3089
    # ideally we should be able to do this, but it's not possible
    # at the time of writing because fhe "filename" is incomplete
    # see the companion comment in load_sample
    # unified_table.set_index("filename", inplace=True)
    # unified_table.index.rename("id", inplace=True)
    return unified_table


def _get_test_data_dir_path():
    p = Path(ytcfg.get("yt", "test_data_dir"))
    if p.is_dir():
        return p
    warn(
        "Storage directory from yt config doesn't exist "
        f"(currently set to '{p}'). "
        "Current working directory will be used instead."
    )
    return Path.cwd()


def lookup_on_disk_data(fn) -> Path:
    """
    Look for data file/dir on disk.

    Returns
    -------

    pathlib.Path to a file/dir matching fn if any

    Raises
    ------
    FileNotFoundError
    """

    path = Path(fn).expanduser().resolve()

    if path.exists():
        return path

    err_msg = f"No such file or directory: '{fn}'."
    test_data_dir = _get_test_data_dir_path()
    if not test_data_dir.is_dir():
        raise FileNotFoundError(err_msg)

    alt_path = _get_test_data_dir_path().joinpath(fn).resolve()
    if alt_path != path:
        if alt_path.exists():
            return alt_path
        err_msg += f"\n(Also tried '{alt_path}')."
    raise FileNotFoundError(err_msg)


@lru_cache(maxsize=128)
def _get_pooch_instance():
    data_registry = get_data_registry_table()
    cache_storage = _get_test_data_dir_path() / "yt_download_cache"

    registry = {k: v["hash"] for k, v in _get_sample_data_registry().items()}
    return pooch.create(
        path=cache_storage, base_url="https://yt-project.org/data/", registry=registry
    )


def _download_sample_data_file(
    filename, progressbar=True, timeout: Optional[int] = None
):
    """
    Download a file by url.

    Returns
    -------
    storage_filename : location of the downloaded file

    """
    downloader = pooch.HTTPDownloader(progressbar=progressbar, timeout=timeout)

    poochie = _get_pooch_instance()
    poochie.fetch(filename, downloader=downloader)
    return Path.joinpath(poochie.path, filename)



from yt.fields.local_fields import add_field
from yt.utilities.linear_interpolators import (
    BilinearFieldInterpolator,
    TrilinearFieldInterpolator,
    UnilinearFieldInterpolator,
)

_int_class = {
    1: UnilinearFieldInterpolator,
    2: BilinearFieldInterpolator,
    3: TrilinearFieldInterpolator,
}


def add_interpolated_field(
    name,
    units,
    table_data,
    axes_data,
    axes_fields,
    ftype="gas",
    particle_type=False,
    validators=None,
    truncate=True,
):
    if len(table_data.shape) not in _int_class:
        raise RuntimeError(
            "Interpolated field can only be created from 1d, 2d, or 3d data."
        )

    if len(axes_fields) != len(axes_data) or len(axes_fields) != len(table_data.shape):
        raise RuntimeError(
            "Data dimension mismatch: data is %d, "
            "%d axes data provided, and %d axes fields provided."
            % (len(table_data.shape), len(axes_data), len(axes_fields))
        )

    int_class = _int_class[len(table_data.shape)]
    my_interpolator = int_class(table_data, axes_data, axes_fields, truncate=truncate)

    def _interpolated_field(field, data):
        return my_interpolator(data)

    add_field(
        (ftype, name),
        function=_interpolated_field,
        units=units,
        validators=validators,
        particle_type=particle_type,
    )



import numpy as np

from yt.utilities.on_demand_imports import _h5py as h5py


def h5rd(fname, path, dtype=None):
    """Read Data. Return a dataset located at <path> in file <fname> as
    a numpy array.
    e.g. rd( fname, '/PartType0/Coordinates' )."""

    data = None
    fid = h5py.h5f.open(fname.encode("latin-1"), h5py.h5f.ACC_RDONLY)
    dg = h5py.h5d.open(fid, path.encode("ascii"))
    if dtype is None:
        dtype = dg.dtype
    data = np.zeros(dg.shape, dtype=dtype)
    dg.read(h5py.h5s.ALL, h5py.h5s.ALL, data)
    fid.close()
    return data


class IonTableSpectrum:

    """A class to handle the HM01 spectra in the OWLS ionization tables."""

    def __init__(self, ion_file):
        where = "/header/spectrum/gammahi"
        self.GH1 = h5rd(ion_file, where)  # GH1[1/s]

        where = "/header/spectrum/logenergy_ryd"
        self.logryd = h5rd(ion_file, where)  # E[ryd]

        where = "/header/spectrum/logflux"
        self.logflux = h5rd(ion_file, where)  # J[ergs/s/Hz/Sr/cm^2]

        where = "/header/spectrum/redshift"
        self.z = h5rd(ion_file, where)  # z

    def return_table_GH1_at_z(self, z):
        # find redshift indices
        # -----------------------------------------------------------------
        i_zlo = np.argmin(np.abs(self.z - z))
        if self.z[i_zlo] < z:
            i_zhi = i_zlo + 1
        else:
            i_zhi = i_zlo
            i_zlo = i_zlo - 1

        z_frac = (z - self.z[i_zlo]) / (self.z[i_zhi] - self.z[i_zlo])

        # find GH1 from table
        # -----------------------------------------------------------------
        logGH1_all = np.log10(self.GH1)
        dlog_GH1 = logGH1_all[i_zhi] - logGH1_all[i_zlo]

        logGH1_table = logGH1_all[i_zlo] + z_frac * dlog_GH1
        GH1_table = 10.0**logGH1_table

        return GH1_table


class IonTableOWLS:

    """A class to handle OWLS ionization tables."""

    DELTA_nH = 0.25
    DELTA_T = 0.1

    def __init__(self, ion_file):
        self.ion_file = ion_file

        # ionbal is indexed like [nH, T, z]
        # nH and T are log quantities
        # ---------------------------------------------------------------
        self.nH = h5rd(ion_file, "/logd")  # log nH [cm^-3]
        self.T = h5rd(ion_file, "/logt")  # log T [K]
        self.z = h5rd(ion_file, "/redshift")  # z

        # read the ionization fractions
        # linear values stored in file so take log here
        # ionbal is the ionization balance (i.e. fraction)
        # ---------------------------------------------------------------
        self.ionbal = h5rd(ion_file, "/ionbal").astype(np.float64)
        self.ionbal_orig = self.ionbal.copy()

        ipositive = self.ionbal > 0.0
        izero = np.logical_not(ipositive)
        self.ionbal[izero] = self.ionbal[ipositive].min()

        self.ionbal = np.log10(self.ionbal)

        # load in background spectrum
        # ---------------------------------------------------------------
        self.spectrum = IonTableSpectrum(ion_file)

        # calculate the spacing along each dimension
        # ---------------------------------------------------------------
        self.dnH = self.nH[1:] - self.nH[0:-1]
        self.dT = self.T[1:] - self.T[0:-1]
        self.dz = self.z[1:] - self.z[0:-1]

        self.order_str = "[log nH, log T, z]"

    # sets iz and fz
    # -----------------------------------------------------
    def set_iz(self, z):
        if z <= self.z[0]:
            self.iz = 0
            self.fz = 0.0
        elif z >= self.z[-1]:
            self.iz = len(self.z) - 2
            self.fz = 1.0
        else:
            for iz in range(len(self.z) - 1):
                if z < self.z[iz + 1]:
                    self.iz = iz
                    self.fz = (z - self.z[iz]) / self.dz[iz]
                    break

    # interpolate the table at a fixed redshift for the input
    # values of nH and T ( input should be log ).  A simple
    # tri-linear interpolation is used.
    # -----------------------------------------------------
    def interp(self, nH, T):
        nH = np.array(nH)
        T = np.array(T)

        if nH.size != T.size:
            raise ValueError(" owls_ion_tables: array size mismatch !!! ")

        # field discovery will have nH.size == 1 and T.size == 1
        # in that case we simply return 1.0

        if nH.size == 1 and T.size == 1:
            ionfrac = 1.0
            return ionfrac

        # find inH and fnH
        # -----------------------------------------------------
        x_nH = (nH - self.nH[0]) / self.DELTA_nH
        x_nH_clip = np.clip(x_nH, 0.0, self.nH.size - 1.001)
        fnH, inH = np.modf(x_nH_clip)
        inH = inH.astype(np.int32)

        # find iT and fT
        # -----------------------------------------------------
        x_T = (T - self.T[0]) / self.DELTA_T
        x_T_clip = np.clip(x_T, 0.0, self.T.size - 1.001)
        fT, iT = np.modf(x_T_clip)
        iT = iT.astype(np.int32)

        # short names for previously calculated iz and fz
        # -----------------------------------------------------
        iz = self.iz
        fz = self.fz

        # calculate interpolated value
        # use tri-linear interpolation on the log values
        # -----------------------------------------------------

        ionfrac = (
            self.ionbal[inH, iT, iz] * (1 - fnH) * (1 - fT) * (1 - fz)
            + self.ionbal[inH + 1, iT, iz] * (fnH) * (1 - fT) * (1 - fz)
            + self.ionbal[inH, iT + 1, iz] * (1 - fnH) * (fT) * (1 - fz)
            + self.ionbal[inH, iT, iz + 1] * (1 - fnH) * (1 - fT) * (fz)
            + self.ionbal[inH + 1, iT, iz + 1] * (fnH) * (1 - fT) * (fz)
            + self.ionbal[inH, iT + 1, iz + 1] * (1 - fnH) * (fT) * (fz)
            + self.ionbal[inH + 1, iT + 1, iz] * (fnH) * (fT) * (1 - fz)
            + self.ionbal[inH + 1, iT + 1, iz + 1] * (fnH) * (fT) * (fz)
        )

        return 10**ionfrac



from unittest import SkipTest

import matplotlib

from yt.testing import requires_backend

active_backend = matplotlib.get_backend()
inactive_backend = ({"gtkagg", "macosx", "wx", "tkagg"} - {active_backend}).pop()


def test_requires_inactive_backend():
    @requires_backend(inactive_backend)
    def foo():
        return

    try:
        foo()
    except SkipTest:
        pass
    else:
        raise AssertionError(
            "@requires_backend appears to be broken (skip was expected)"
        )


def test_requires_active_backend():
    @requires_backend(active_backend)
    def foo():
        return

    try:
        foo()
    except SkipTest:
        raise AssertionError(
            "@requires_backend appears to be broken (skip was not expected)"
        ) from None



# This handles the command line.

import argparse
import os
import signal
import sys
from typing import List

from yt.config import ytcfg
from yt.funcs import (
    mylog,
    paste_traceback,
    paste_traceback_detailed,
    signal_ipython,
    signal_print_traceback,
)
from yt.utilities import rpdb

exe_name = os.path.basename(sys.executable)


# At import time, we determined whether or not we're being run in parallel.
def turn_on_parallelism():
    parallel_capable = False
    try:
        # we import this to check if mpi4py is installed
        from mpi4py import MPI  # NOQA
    except ImportError as e:
        mylog.error(
            "Warning: Attempting to turn on parallelism, "
            "but mpi4py import failed. Try pip install mpi4py."
        )
        raise e
        # Now we have to turn on the parallelism from the perspective of the
    # parallel_analysis_interface
    from yt.utilities.parallel_tools.parallel_analysis_interface import (
        enable_parallelism,
    )

    parallel_capable = enable_parallelism()
    return parallel_capable


# This fallback is for Paraview:

# We use two signals, SIGUSR1 and SIGUSR2.  In a non-threaded environment,
# we set up handlers to process these by printing the current stack and to
# raise a RuntimeError.  The latter can be used, inside pdb, to catch an error
# and then examine the current stack.
try:
    signal.signal(signal.SIGUSR1, signal_print_traceback)
    mylog.debug("SIGUSR1 registered for traceback printing")
    signal.signal(signal.SIGUSR2, signal_ipython)
    mylog.debug("SIGUSR2 registered for IPython Insertion")
except (ValueError, RuntimeError, AttributeError):  # Not in main thread
    pass


class SetExceptionHandling(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        # If we recognize one of the arguments on the command line as indicating a
        # different mechanism for handling tracebacks, we attach one of those handlers
        # and remove the argument from sys.argv.
        #
        if self.dest == "paste":
            sys.excepthook = paste_traceback
            mylog.debug("Enabling traceback pasting")
        elif self.dest == "paste-detailed":
            sys.excepthook = paste_traceback_detailed
            mylog.debug("Enabling detailed traceback pasting")
        elif self.dest == "detailed":
            import cgitb

            cgitb.enable(format="text")
            mylog.debug("Enabling detailed traceback reporting")
        elif self.dest == "rpdb":
            sys.excepthook = rpdb.rpdb_excepthook
            mylog.debug("Enabling remote debugging")


class SetConfigOption(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        param, val = values.split("=")
        mylog.debug("Overriding config: %s = %s", param, val)
        ytcfg["yt", param] = val
        if param == "log_level":  # special case
            mylog.setLevel(int(val))


class YTParser(argparse.ArgumentParser):
    def error(self, message):
        """error(message: string)

        Prints a help message that is more detailed than the argparse default
        and then exits.
        """
        self.print_help(sys.stderr)
        self.exit(2, f"{self.prog}: error: {message}\n")


parser = YTParser(description="yt command line arguments")
parser.add_argument(
    "--config",
    action=SetConfigOption,
    help="Set configuration option, in the form param=value",
)
parser.add_argument(
    "--paste",
    action=SetExceptionHandling,
    help="Paste traceback to paste.yt-project.org",
    nargs=0,
)
parser.add_argument(
    "--paste-detailed",
    action=SetExceptionHandling,
    help="Paste a detailed traceback with local variables to " + "paste.yt-project.org",
    nargs=0,
)
parser.add_argument(
    "--detailed",
    action=SetExceptionHandling,
    help="Display detailed traceback.",
    nargs=0,
)
parser.add_argument(
    "--rpdb",
    action=SetExceptionHandling,
    help="Enable remote pdb interaction (for parallel debugging).",
    nargs=0,
)
parser.add_argument(
    "--parallel",
    action="store_true",
    default=False,
    dest="parallel",
    help="Run in MPI-parallel mode (must be launched as an MPI task)",
)
if not hasattr(sys, "argv") or sys.argv is None:
    sys.argv = []

unparsed_args: List[str] = []

parallel_capable = False
if not ytcfg.get("yt", "internals", "command_line"):
    opts, unparsed_args = parser.parse_known_args()
    # THIS IS NOT SUCH A GOOD IDEA:
    # sys.argv = [a for a in unparsed_args]
    if opts.parallel:
        parallel_capable = turn_on_parallelism()
    subparsers = parser.add_subparsers(
        title="subcommands",
        dest="subcommands",
        description="Valid subcommands",
    )
else:
    subparsers = parser.add_subparsers(
        title="subcommands",
        dest="subcommands",
        description="Valid subcommands",
    )

    def print_help(*args, **kwargs):
        parser.print_help()

    help_parser = subparsers.add_parser("help", help="Print help message")
    help_parser.set_defaults(func=print_help)


if parallel_capable:
    pass
elif (
    exe_name in ["mpi4py", "embed_enzo", "python{}.{}-mpi".format(*sys.version_info)]
    or "_parallel" in dir(sys)
    or any("ipengine" in arg for arg in sys.argv)
    or any("cluster-id" in arg for arg in sys.argv)
):
    parallel_capable = turn_on_parallelism()
else:
    parallel_capable = False



import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

import yt
from yt import particle_filter
from yt.testing import (
    assert_rel_equal,
    fake_particle_ds,
    fake_random_ds,
    fake_sph_orientation_ds,
    requires_file,
)


def setup():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_extrema():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(
            16,
            nprocs=nprocs,
            fields=("density", "velocity_x", "velocity_y", "velocity_z"),
            units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
        )
        for sp in [ds.sphere("c", (0.25, "unitary")), ds.r[0.5, :, :]]:
            mi, ma = sp.quantities["Extrema"](("gas", "density"))
            assert_equal(mi, np.nanmin(sp[("gas", "density")]))
            assert_equal(ma, np.nanmax(sp[("gas", "density")]))
            dd = ds.all_data()
            mi, ma = dd.quantities["Extrema"](("gas", "density"))
            assert_equal(mi, np.nanmin(dd[("gas", "density")]))
            assert_equal(ma, np.nanmax(dd[("gas", "density")]))
            sp = ds.sphere("max", (0.25, "unitary"))
            assert_equal(np.any(np.isnan(sp[("gas", "radial_velocity")])), False)
            mi, ma = dd.quantities["Extrema"](("gas", "radial_velocity"))
            assert_equal(mi, np.nanmin(dd[("gas", "radial_velocity")]))
            assert_equal(ma, np.nanmax(dd[("gas", "radial_velocity")]))


def test_average():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs=nprocs, fields=("density",), units=("g/cm**3",))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:
            my_mean = ad.quantities["WeightedAverageQuantity"](
                ("gas", "density"), ("index", "ones")
            )
            assert_rel_equal(my_mean, ad[("gas", "density")].mean(), 12)

            my_mean = ad.quantities["WeightedAverageQuantity"](
                ("gas", "density"), ("gas", "cell_mass")
            )
            a_mean = (ad[("gas", "density")] * ad[("gas", "cell_mass")]).sum() / ad[
                ("gas", "cell_mass")
            ].sum()
            assert_rel_equal(my_mean, a_mean, 12)


def test_standard_deviation():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs=nprocs, fields=("density",), units=("g/cm**3",))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:
            my_std, my_mean = ad.quantities["WeightedStandardDeviation"](
                ("gas", "density"), ("index", "ones")
            )
            assert_rel_equal(my_mean, ad[("gas", "density")].mean(), 12)
            assert_rel_equal(my_std, ad[("gas", "density")].std(), 12)

            my_std, my_mean = ad.quantities["WeightedStandardDeviation"](
                ("gas", "density"), ("gas", "cell_mass")
            )
            a_mean = (ad[("gas", "density")] * ad[("gas", "cell_mass")]).sum() / ad[
                ("gas", "cell_mass")
            ].sum()
            assert_rel_equal(my_mean, a_mean, 12)
            a_std = np.sqrt(
                (
                    ad[("gas", "cell_mass")] * (ad[("gas", "density")] - a_mean) ** 2
                ).sum()
                / ad[("gas", "cell_mass")].sum()
            )
            assert_rel_equal(my_std, a_std, 12)


def test_max_location():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs=nprocs, fields=("density",), units=("g/cm**3",))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:
            mv, x, y, z = ad.quantities.max_location(("gas", "density"))

            assert_equal(mv, ad[("gas", "density")].max())

            mi = np.argmax(ad[("gas", "density")])

            assert_equal(ad[("index", "x")][mi], x)
            assert_equal(ad[("index", "y")][mi], y)
            assert_equal(ad[("index", "z")][mi], z)


def test_min_location():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs=nprocs, fields=("density",), units=("g/cm**3",))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:
            mv, x, y, z = ad.quantities.min_location(("gas", "density"))

            assert_equal(mv, ad[("gas", "density")].min())

            mi = np.argmin(ad[("gas", "density")])

            assert_equal(ad[("index", "x")][mi], x)
            assert_equal(ad[("index", "y")][mi], y)
            assert_equal(ad[("index", "z")][mi], z)


def test_sample_at_min_field_values():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(
            16,
            nprocs=nprocs,
            fields=("density", "temperature", "velocity_x"),
            units=("g/cm**3", "K", "cm/s"),
        )
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:
            mv, temp, vm = ad.quantities.sample_at_min_field_values(
                ("gas", "density"), [("gas", "temperature"), ("gas", "velocity_x")]
            )

            assert_equal(mv, ad[("gas", "density")].min())

            mi = np.argmin(ad[("gas", "density")])

            assert_equal(ad[("gas", "temperature")][mi], temp)
            assert_equal(ad[("gas", "velocity_x")][mi], vm)


def test_sample_at_max_field_values():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(
            16,
            nprocs=nprocs,
            fields=("density", "temperature", "velocity_x"),
            units=("g/cm**3", "K", "cm/s"),
        )
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:
            mv, temp, vm = ad.quantities.sample_at_max_field_values(
                ("gas", "density"), [("gas", "temperature"), ("gas", "velocity_x")]
            )

            assert_equal(mv, ad[("gas", "density")].max())

            mi = np.argmax(ad[("gas", "density")])

            assert_equal(ad[("gas", "temperature")][mi], temp)
            assert_equal(ad[("gas", "velocity_x")][mi], vm)


def test_in_memory_sph_derived_quantities():
    ds = fake_sph_orientation_ds()
    ad = ds.all_data()

    ang_mom = ad.quantities.angular_momentum_vector()
    assert_equal(ang_mom, [0, 0, 0])

    bv = ad.quantities.bulk_velocity()
    assert_equal(bv, [0, 0, 0])

    com = ad.quantities.center_of_mass()
    assert_equal(com, [1 / 7, (1 + 2) / 7, (1 + 2 + 3) / 7])

    ex = ad.quantities.extrema([("io", "x"), ("io", "y"), ("io", "z")])
    for fex, ans in zip(ex, [[0, 1], [0, 2], [0, 3]]):
        assert_equal(fex, ans)

    for d, v, l in zip("xyz", [1, 2, 3], [[1, 0, 0], [0, 2, 0], [0, 0, 3]]):
        max_d, x, y, z = ad.quantities.max_location(("io", d))
        assert_equal(max_d, v)
        assert_equal([x, y, z], l)

    for d in "xyz":
        min_d, x, y, z = ad.quantities.min_location(("io", d))
        assert_equal(min_d, 0)
        assert_equal([x, y, z], [0, 0, 0])

    tot_m = ad.quantities.total_mass()
    assert_equal(tot_m, [7, 0])

    weighted_av_z = ad.quantities.weighted_average_quantity(("io", "z"), ("io", "z"))
    assert_equal(weighted_av_z, 7 / 3)


iso_collapse = "IsothermalCollapse/snap_505"
tipsy_gal = "TipsyGalaxy/galaxy.00300"


@requires_file(iso_collapse)
@requires_file(tipsy_gal)
def test_sph_datasets_derived_quantities():
    for fname in [tipsy_gal, iso_collapse]:
        ds = yt.load(fname)
        ad = ds.all_data()
        use_particles = "nbody" in ds.particle_types
        ad.quantities.angular_momentum_vector()
        ad.quantities.bulk_velocity(True, use_particles)
        ad.quantities.center_of_mass(True, use_particles)
        ad.quantities.extrema([("gas", "density"), ("gas", "temperature")])
        ad.quantities.min_location(("gas", "density"))
        ad.quantities.max_location(("gas", "density"))
        ad.quantities.total_mass()
        ad.quantities.weighted_average_quantity(("gas", "density"), ("gas", "mass"))


def test_derived_quantities_with_particle_types():
    ds = fake_particle_ds()

    @particle_filter(requires=["particle_position_x"], filtered_type="all")
    def low_x(pfilter, data):
        return (
            data[(pfilter.filtered_type, "particle_position_x")].in_units("code_length")
            < 0.5
        )

    ds.add_particle_filter("low_x")

    ad = ds.all_data()

    for ptype in ["all", "low_x"]:
        # Check bulk velocity
        bulk_vx = (
            ad[(ptype, "particle_mass")]
            * ad[(ptype, "particle_velocity_x")]
            / ad[(ptype, "particle_mass")].sum()
        ).sum()
        assert_almost_equal(
            ad.quantities.bulk_velocity(
                use_gas=False, use_particles=True, particle_type=ptype
            )[0],
            bulk_vx,
            5,
        )

        # Check center of mass
        com_x = (
            ad[(ptype, "particle_mass")]
            * ad[(ptype, "particle_position_x")]
            / ad[(ptype, "particle_mass")].sum()
        ).sum()
        assert_almost_equal(
            ad.quantities.center_of_mass(
                use_gas=False, use_particles=True, particle_type=ptype
            )[0],
            com_x,
            5,
        )

        # Check angular momentum vector
        l_x = (
            ad[(ptype, "particle_specific_angular_momentum_x")]
            * ad[(ptype, "particle_mass")]
            / ad[(ptype, "particle_mass")].sum()
        ).sum()
        assert_almost_equal(
            ad.quantities.angular_momentum_vector(
                use_gas=False, use_particles=True, particle_type=ptype
            )[0],
            l_x,
            5,
        )

    # Check spin parameter values
    assert_almost_equal(
        ad.quantities.spin_parameter(use_gas=False, use_particles=True),
        655.7311454765503,
    )
    assert_almost_equal(
        ad.quantities.spin_parameter(
            use_gas=False, use_particles=True, particle_type="low_x"
        ),
        1309.164886405665,
    )



import errno
import glob
import os
import shutil
import subprocess
import tempfile
import time
import uuid

from docutils import nodes
from docutils.parsers.rst import Directive


class PythonScriptDirective(Directive):
    """Execute an inline python script and display images.

    This uses exec to execute an inline python script, copies
    any images produced by the script, and embeds them in the document
    along with the script.

    """

    required_arguments = 0
    optional_arguments = 0
    has_content = True

    def run(self):
        cwd = os.getcwd()
        tmpdir = tempfile.mkdtemp()
        os.chdir(tmpdir)

        rst_file = self.state_machine.document.attributes["source"]
        rst_dir = os.path.abspath(os.path.dirname(rst_file))

        image_dir, image_rel_dir = make_image_dir(setup, rst_dir)

        # Construct script from cell content
        content = "\n".join(self.content)
        with open("temp.py", "w") as f:
            f.write(content)

        # Use sphinx logger?
        uid = uuid.uuid4().hex[:8]
        print("")
        print(f">> Contents of the script: {uid}")
        print(content)
        print("")

        start = time.time()
        subprocess.call(["python", "temp.py"])
        print(f">> The execution of the script {uid} took {time.time() - start:f} s")
        text = ""
        for im in sorted(glob.glob("*.png")):
            text += get_image_tag(im, image_dir, image_rel_dir)

        code = content

        literal = nodes.literal_block(code, code)
        literal["language"] = "python"

        attributes = {"format": "html"}
        img_node = nodes.raw("", text, **attributes)

        # clean up
        os.chdir(cwd)
        shutil.rmtree(tmpdir, True)

        return [literal, img_node]


def setup(app):
    app.add_directive("python-script", PythonScriptDirective)
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    retdict = dict(version="0.1", parallel_read_safe=True, parallel_write_safe=True)

    return retdict


def get_image_tag(filename, image_dir, image_rel_dir):
    my_uuid = uuid.uuid4().hex
    shutil.move(filename, image_dir + os.path.sep + my_uuid + filename)
    relative_filename = image_rel_dir + os.path.sep + my_uuid + filename
    return f'<img src="{relative_filename}" width="600"><br>'


def make_image_dir(setup, rst_dir):
    image_dir = setup.app.builder.outdir + os.path.sep + "_images"
    rel_dir = os.path.relpath(setup.confdir, rst_dir)
    image_rel_dir = rel_dir + os.path.sep + "_images"
    thread_safe_mkdir(image_dir)
    return image_dir, image_rel_dir


def thread_safe_mkdir(dirname):
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise



import yt

# Load the dataset.
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Create a sphere of radius 100 kpc in the center of the domain.
my_sphere = ds.sphere("c", (100.0, "kpc"))

# Create a PhasePlot object.
# Setting weight to None will calculate a sum.
# Setting weight to a field will calculate an average
# weighted by that field.
plot = yt.PhasePlot(
    my_sphere,
    ("gas", "density"),
    ("gas", "temperature"),
    ("gas", "mass"),
    weight_field=None,
)

# Set the units of mass to be in solar masses (not the default in cgs)
plot.set_unit(("gas", "mass"), "Msun")

# Save the image.
# Optionally, give a string as an argument
# to name files with a keyword.
plot.save()



import numpy as np

import yt
from yt.units import kpc
from yt.visualization.volume_rendering.api import PointSource

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

sc = yt.create_scene(ds)

np.random.seed(1234567)

npoints = 1000

# Random particle positions
vertices = np.random.random([npoints, 3]) * 200 * kpc

# Random colors
colors = np.random.random([npoints, 4])

# Set alpha value to something that produces a good contrast with the volume
# rendering
colors[:, 3] = 0.1

points = PointSource(vertices, colors=colors)
sc.add_source(points)

sc.camera.width = 300 * kpc

sc.save(sigma_clip=5)



from functools import cached_property
from itertools import groupby

import numpy as np

from yt.geometry.selection_routines import AlwaysSelector
from yt.utilities.io_handler import BaseIOHandler


# http://stackoverflow.com/questions/2361945/detecting-consecutive-integers-in-a-list
def particle_sequences(grids):
    g_iter = sorted(grids, key=lambda g: g.id)
    for _k, g in groupby(enumerate(g_iter), lambda i_x: i_x[0] - i_x[1].id):
        seq = [v[1] for v in g]
        yield seq[0], seq[-1]


def grid_sequences(grids):
    g_iter = sorted(grids, key=lambda g: g.id)
    for _k, g in groupby(enumerate(g_iter), lambda i_x1: i_x1[0] - i_x1[1].id):
        seq = [v[1] for v in g]
        yield seq


def determine_particle_fields(handle):
    try:
        particle_fields = [
            s[0].decode("ascii", "ignore").strip() for s in handle["/particle names"][:]
        ]
        _particle_fields = {"particle_" + s: i for i, s in enumerate(particle_fields)}
    except KeyError:
        _particle_fields = {}
    return _particle_fields


def _conditionally_split_arrays(inp_arrays, condition):
    output_true = []
    output_false = []
    for arr in inp_arrays:
        output_true.append(arr[condition])
        output_false.append(arr[~condition])
    return output_true, output_false


class IOHandlerFLASH(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "flash_hdf5"

    def __init__(self, ds):
        super().__init__(ds)
        # Now we cache the particle fields
        self._handle = ds._handle
        self._particle_handle = ds._particle_handle
        self._particle_fields = determine_particle_fields(self._particle_handle)

    def _read_particles(
        self, fields_to_read, type, args, grid_list, count_list, conv_factors
    ):
        pass

    def io_iter(self, chunks, fields):
        f = self._handle
        for chunk in chunks:
            for field in fields:
                # Note that we *prefer* to iterate over the fields on the
                # outside; here, though, we're iterating over them on the
                # inside because we may exhaust our chunks.
                ftype, fname = field
                ds = f[f"/{fname}"]
                for gs in grid_sequences(chunk.objs):
                    start = gs[0].id - gs[0]._id_offset
                    end = gs[-1].id - gs[-1]._id_offset + 1
                    data = ds[start:end, :, :, :]
                    for i, g in enumerate(gs):
                        yield field, g, self._read_obj_field(g, field, (data, i))

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        f_part = self._particle_handle
        p_ind = self.ds.index._particle_indices
        px, py, pz = (self._particle_fields[f"particle_pos{ax}"] for ax in "xyz")
        pblk = self._particle_fields["particle_blk"]
        blockless_buffer = self.ds.index._blockless_particle_count
        p_fields = f_part["/tracer particles"]
        assert len(ptf) == 1
        ptype = list(ptf.keys())[0]
        bxyz = []
        # We need to track all the particles that don't have blocks and make
        # sure they get yielded too.  But, we also want our particles to largely
        # line up with the grids they are resident in.  So we keep a buffer of
        # particles.  That buffer is checked for any "blockless" particles,
        # which get yielded at the end.
        for chunk in chunks:
            start = end = None
            for g1, g2 in particle_sequences(chunk.objs):
                start = p_ind[g1.id - g1._id_offset]
                end_nobuff = p_ind[g2.id - g2._id_offset + 1]
                end = end_nobuff + blockless_buffer
                maxlen = end_nobuff - start
                blk = np.asarray(p_fields[start:end, pblk], dtype="=i8") >= 0
                # Can we use something like np.choose here?
                xyz = [
                    np.asarray(p_fields[start:end, _], dtype="=f8")
                    for _ in (px, py, pz)
                ]
                (x, y, z), _xyz = _conditionally_split_arrays(xyz, blk)
                if _xyz[0].size > 0:
                    bxyz.append(_xyz)
                yield ptype, (x[:maxlen], y[:maxlen], z[:maxlen]), 0.0
        if len(bxyz) == 0:
            return
        bxyz = np.concatenate(bxyz, axis=-1)
        yield ptype, (bxyz[0, :], bxyz[1, :], bxyz[2, :]), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        f_part = self._particle_handle
        p_ind = self.ds.index._particle_indices
        px, py, pz = (self._particle_fields[f"particle_pos{ax}"] for ax in "xyz")
        pblk = self._particle_fields["particle_blk"]
        blockless_buffer = self.ds.index._blockless_particle_count
        p_fields = f_part["/tracer particles"]
        assert len(ptf) == 1
        ptype = list(ptf.keys())[0]
        field_list = ptf[ptype]
        bxyz = []
        bfields = {(ptype, field): [] for field in field_list}
        for chunk in chunks:
            for g1, g2 in particle_sequences(chunk.objs):
                start = p_ind[g1.id - g1._id_offset]
                end_nobuff = p_ind[g2.id - g2._id_offset + 1]
                end = end_nobuff + blockless_buffer
                maxlen = end_nobuff - start
                blk = np.asarray(p_fields[start:end, pblk], dtype="=i8") >= 0
                xyz = [
                    np.asarray(p_fields[start:end, _], dtype="=f8")
                    for _ in (px, py, pz)
                ]
                (x, y, z), _xyz = _conditionally_split_arrays(xyz, blk)
                x, y, z = (_[:maxlen] for _ in (x, y, z))
                if _xyz[0].size > 0:
                    bxyz.append(_xyz)
                mask = selector.select_points(x, y, z, 0.0)
                blockless = (_xyz[0] > 0).any()
                # This checks if none of the particles within these blocks are
                # included in the mask We need to also allow for blockless
                # particles to be selected.
                if mask is None and not blockless:
                    continue
                for field in field_list:
                    fi = self._particle_fields[field]
                    (data,), (bdata,) = _conditionally_split_arrays(
                        [p_fields[start:end, fi]], blk
                    )
                    # Note that we have to apply mask to the split array, since
                    # we select on the split array.
                    if mask is not None:
                        # Be sure to truncate off the buffer length!!
                        yield (ptype, field), data[:maxlen][mask]
                    if blockless:
                        bfields[ptype, field].append(bdata)
        if len(bxyz) == 0:
            return
        bxyz = np.concatenate(bxyz, axis=-1)
        mask = selector.select_points(bxyz[0, :], bxyz[1, :], bxyz[2, :], 0.0)
        if mask is None:
            return
        for field in field_list:
            data_field = np.concatenate(bfields[ptype, field], axis=-1)
            yield (ptype, field), data_field[mask]

    def _read_obj_field(self, obj, field, ds_offset=None):
        if ds_offset is None:
            ds_offset = (None, -1)
        ds, offset = ds_offset
        # our context here includes datasets and whatnot that are opened in the
        # hdf5 file
        if ds is None:
            ds = self._handle[f"/{field[1]}"]
        if offset == -1:
            data = ds[obj.id - obj._id_offset, :, :, :].transpose()
        else:
            data = ds[offset, :, :, :].transpose()
        return data.astype("=f8")

    def _read_chunk_data(self, chunk, fields):
        f = self._handle
        rv = {}
        for g in chunk.objs:
            rv[g.id] = {}
        # Split into particles and non-particles
        fluid_fields, particle_fields = [], []
        for ftype, fname in fields:
            if ftype in self.ds.particle_types:
                particle_fields.append((ftype, fname))
            else:
                fluid_fields.append((ftype, fname))
        if len(particle_fields) > 0:
            selector = AlwaysSelector(self.ds)
            rv.update(self._read_particle_selection([chunk], selector, particle_fields))
        if len(fluid_fields) == 0:
            return rv
        for field in fluid_fields:
            ftype, fname = field
            ds = f[f"/{fname}"]
            for gs in grid_sequences(chunk.objs):
                start = gs[0].id - gs[0]._id_offset
                end = gs[-1].id - gs[-1]._id_offset + 1
                data = ds[start:end, :, :, :].transpose()
                for i, g in enumerate(gs):
                    rv[g.id][field] = np.asarray(data[..., i], "=f8")
        return rv


class IOHandlerFLASHParticle(BaseIOHandler):
    _particle_reader = True
    _dataset_type = "flash_particle_hdf5"

    def __init__(self, ds):
        super().__init__(ds)
        # Now we cache the particle fields
        self._handle = ds._handle
        self._particle_fields = determine_particle_fields(self._handle)
        self._position_fields = [
            self._particle_fields[f"particle_pos{ax}"] for ax in "xyz"
        ]

    @property
    def chunksize(self):
        return 32**3

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        data_files = set()
        assert len(ptf) == 1
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        px, py, pz = self._position_fields
        p_fields = self._handle["/tracer particles"]
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            pxyz = np.asarray(
                p_fields[data_file.start : data_file.end, [px, py, pz]], dtype="=f8"
            )
            yield "io", pxyz.T, 0.0

    def _yield_coordinates(self, data_file, needed_ptype=None):
        px, py, pz = self._position_fields
        p_fields = self._handle["/tracer particles"]
        pxyz = np.asarray(
            p_fields[data_file.start : data_file.end, [px, py, pz]], dtype="=f8"
        )
        yield ("io", pxyz)

    def _read_particle_data_file(self, data_file, ptf, selector=None):
        px, py, pz = self._position_fields
        p_fields = self._handle["/tracer particles"]
        si, ei = data_file.start, data_file.end

        data_return = {}
        # This should just be a single item
        for ptype, field_list in sorted(ptf.items()):
            x = np.asarray(p_fields[si:ei, px], dtype="=f8")
            y = np.asarray(p_fields[si:ei, py], dtype="=f8")
            z = np.asarray(p_fields[si:ei, pz], dtype="=f8")
            if selector:
                mask = selector.select_points(x, y, z, 0.0)
            del x, y, z
            if mask is None:
                continue

            for field in field_list:
                fi = self._particle_fields[field]
                data = p_fields[si:ei, fi]
                if selector:
                    data = data[mask]
                data_return[(ptype, field)] = data

        return data_return

    def _read_particle_fields(self, chunks, ptf, selector):
        assert len(ptf) == 1
        yield from super()._read_particle_fields(chunks, ptf, selector)

    @cached_property
    def _pcount(self):
        return self._handle["/localnp"][:].sum()

    def _count_particles(self, data_file):
        si, ei = data_file.start, data_file.end
        pcount = self._pcount
        if None not in (si, ei):
            pcount = np.clip(pcount - si, 0, ei - si)
        return {"io": pcount}

    def _identify_fields(self, data_file):
        fields = [("io", field) for field in self._particle_fields]
        return fields, {}
